/*
 * ARDataSource: comments by rdb 02/09/12
 * 
 * History (recent)
 * 02/09/12 rdb Add writeARfdl method; it seems like the only thing in ARBuilder
 *              that is needed when the ARDataSource is not built by ARBuilder
 *              It makes sense to do it here, not there.
 */

package ar;

import edu.unh.sdb.datasource.BasicBlock;
import edu.unh.sdb.datasource.DataBlock;
import edu.unh.sdb.datasource.DataCollection;
import edu.unh.sdb.datasource.DataSource;
import edu.unh.sdb.datasource.FDLDescriptor;
import edu.unh.sdb.datasource.ISBounds;
import error.Coords;
import error.DataSetWriter;
import error.UniformCoords;
import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;

/**
 * Maintain the ARNode tree and the references to the original input
 * datasources that are needed for an AR datasource.
 * 
 * @author svohr
 */
public class ARDataSource 
{    
   //--------------- instance variables --------------------------------
    protected ARNode root;
    
    private int resolutions;
    private int errorOffset;
    
    private ISBounds bounds;
    
    // Data and Error files for the ARDataSource
    private DataBlock [] dataSources;
    private DataBlock [] errorSources;
    
    // Names of Data and Error FDL files
    private String [] dataFDLNames;
    private String [] errorFDLNames;

    // An FDLDescriptor to keep around for reference.
    private FDLDescriptor fdl;
    
    private Coords coords = new UniformCoords();

    //------------------------- constructors ---------------------------
    public ARDataSource ()
    {
        
    }

    public ARDataSource( MRHierarchy mr )
    {
        this.dataSources = mr.getData();
        this.errorSources = mr.getError();
        this.dataFDLNames = mr.getDataFDLNames();
        this.errorFDLNames = mr.getErrorFDLNames();
        
        this.resolutions = dataSources.length;
        this.errorOffset = mr.getErrorOffset();
        
        this.bounds = dataSources[0].copyBounds();
        
        // build base coordinate bounds for the dataSources set.
        bounds.setLower(bounds.getLower().scale(2));
        bounds.setUpper(bounds.getUpper().scale(2));
        
        root = new ARNode( bounds );
        
        root.buildTree( resolutions-1, errorOffset, dataSources, errorSources );
        
        fdl = new FDLDescriptor();
        fdl.setDataFileURL( dataFDLNames[0] );
    }

    /**
     * 
     * @param dataNames
     * @param errorNames 
     */
    public ARDataSource( String[] dataNames, String[] errorNames )
    {
        this.resolutions = dataNames.length;
        
        this.dataSources = new BasicBlock [resolutions];
        this.errorSources = new BasicBlock [resolutions];
        
        this.dataFDLNames = dataNames;
        this.errorFDLNames = errorNames;
        
        /*
         //DB: This will be passed in, no need to read the data sources
         * 
         *rdb This constructor is used by the code (Dave Bochichio) that is 
         * supposed to run during simulation; I'm a little worried that it 
         * might be used somewhere else where it expects the data sources
         * to be created here.
        
        // open all of the dataSources sources
        for ( int i=0; i < dataNames.length; i++ ) {
            dataSources[i] = DataSource.create("data_" + i, dataNames[i]);
            dataSources[i].activate();
        }
        */
        
        //DB: The data will be passed in, no need to read the data sources
        /* open all of the errorSources sources
        
        for ( int i=1; i < errorNames.length; i++ ) {
            errorSources[i] = DataSource.create("error_" + i, errorNames[i]);
            errorSources[i].activate();
        } */// (remember errorSources[0] is left empty; no errorSources for original)
        
        this.bounds = dataSources[0].copyBounds();
        
        // build base coordinate bounds for the dataSources set.
        bounds.setLower( bounds.getLower().scale(2) );
        bounds.setUpper( bounds.getUpper().scale(2) );
        
        root = new ARNode( bounds );
        fdl = FDLDescriptor.create( dataFDLNames[0] );
        root.buildTree( resolutions-1, dataSources, errorSources );
        
        fdl = FDLDescriptor.create( dataFDLNames[0] );
    }
    
    public ARNode getRoot() {
        return root;
    }

    public void setRoot(ARNode root) 
    {
        this.root = root;
    }

    public DataBlock[] getDataSources() {
        return dataSources;
    }
        
    public DataBlock getDataSource(int index) {
        return dataSources[index];
    }
    
    public void setDataSource(DataBlock d, int index) {
        dataSources[index] = d;
    }

    public DataBlock[] getErrorSources() {
        return errorSources;
    }

    public String[] getDataFDLNames() {
        return dataFDLNames;
    }
    
    public void setDataFDLNames(String s, int index) {
        dataFDLNames[index] = s;
    }

    public String[] getErrorFDLNames() {
        return errorFDLNames;
    }
    
    public void setErrorFDLNames(String s, int index) {
        dataFDLNames[index] = s;
    }

    public int getResolutions() {
        return resolutions;
    }

    public void setResolutions(int resolutions) {
        this.resolutions = resolutions;
        dataFDLNames = new String [resolutions];
        dataSources = new BasicBlock [resolutions];
        errorFDLNames = new String [resolutions];
        errorSources = new BasicBlock [resolutions];
    }
    

    public ISBounds getBounds() {
        return bounds;
    }
    
    public void setBounds(ISBounds i) {
        bounds = i;
    }
    
    public void dump()
    {
        if (root == null) 
            System.err.println ("empty");
        else
            root.dumpAll("");
    }
    
    
    private DataBlock subblock ( ARNodeBlock b )
    {
        int res = b.getResolution();       
        return this.dataSources[ res ].subblock( b.getBounds() );
    }
    
    public void writeARfdl( String outFileName )
    {
      try
      {
         ARWriter out = new ARWriter();
         if ( outFileName != null && outFileName.length() > 0 )
            out.setOutputFile( outFileName );

         out.writeAR( this );
         out.close();
      }
      catch ( IOException e )
      {
         System.err.println( e.getMessage() );
      }

    }
    
    public void writeBlocks ( ARError tolerance )
    {
       dumpBlocks( tolerance );
    }
    public void dumpBlocks ( ARError tolerance )
    {
        
        String parentDir = "ARBlocks";
        System.out.println( parentDir );
        // make a directory for all of the blocks.
        File f = new File( parentDir );
        
        if (! f.mkdir()) 
        {
            System.err.println("Directory existed - deleting and recreating.");
            f.delete();
            f.mkdir();            
        }

        List <ARNodeBlock> blocks = new ArrayList <ARNodeBlock>();
        //root.getLeafBlocks(blocks, tolerance);
        root.prune(tolerance);
        root.condense();
        
        root.getLeafBlocks(blocks);
        
        int spots = Integer.toString( blocks.size() ).length();
        int blockNum = 1;
        
        String dirFormat = String.format("%s/b%%0%dd/", parentDir, spots);
        
        for ( ARNodeBlock b : blocks ) {
            String blockDir = String.format(dirFormat, blockNum);
            File fb = new File( blockDir );
            if ( ! fb.mkdir() ) {
            	System.err.println("Error creating directory for subblock.");
            	fb.delete();
            	fb.mkdir();
            }
            
            dumpBlock (blockDir, b);
            
            blockNum += 1;
        }
        
    }
    
    private void dumpBlock ( String path, ARNodeBlock b )
    {
        String dname = "data.bin";
        String pname = "points.grid";
    
        //fdl = new FDLDescriptor();
        FDLDescriptor fd = (FDLDescriptor)fdl.clone();
        fd.setBounds( b.getBounds() );
        fd.setDataFileURL( dname );
        
//        System.err.println ( path + ": res" 
//                + b.getResolution() + " => " 
//                + b.getBounds().toString() + " = " + b.getBounds().volume() );
        
        // Write the data out.
        try {
            DataSetWriter dw = new DataSetWriter (path + dname, fd);
            
            dw.suppressFDL(true);
            
            DataBlock data = this.subblock ( b );
            System.out.println("Writing size: " + data.volume());
            dw.write(data);
            
            dw.close();
        }
        catch ( IOException e ) {
            System.err.println ( "Error writing data file: " + e.getMessage() );
        }
        
        // write out the coordinates of the points (base)
        ISBounds points = b.getBounds();
        int res = b.getResolution();
        
        try {
            PrintWriter pw = new PrintWriter (new File (path + pname));
            
            // print headers
            pw.println("DATASET RECTILINEAR_GRID");
            
            // print dimension information
            pw.print("DIMENSIONS ");
            for ( int i = points.dim()-1; i >=0; i-- ) { 
                pw.print( points.getDimension(i) );
                if ( i > 0 ) 
                    pw.print(" ");
            }
            pw.println();
            
            String labels = "ZYX";
            for ( int i = points.dim()-1; i >=0; i-- ) {
                int start = points.getLower(i);
                int end = points.getUpper(i);

                pw.printf(  "%c_COORDINATES %d int\n", 
                            labels.charAt(i), 
                            points.getDimension(i) );
                
                for ( int j=start; j <= end; j++ ) {
                    int base = coords.lowResToBase1d(res, j);
                    
                    pw.print(base);
                    
                    if (j != end)
                        pw.print(" ");
                }
                
                pw.println();
            }
            
            pw.close();
        }
        catch ( Exception e )
        {
            System.err.println ( "Error writing points file: " + e.getMessage() );
        }
    }


}
