/*
 * ARNode.java: comments by rdb 02/09/12
 */

package ar;


import edu.unh.sdb.datasource.DataBlock;
import edu.unh.sdb.datasource.DataCollection;
import edu.unh.sdb.datasource.DataSource;
import edu.unh.sdb.datasource.ISBounds;
import edu.unh.sdb.datasource.IndexSpaceID;
import java.util.List;

/**
 * Represent a node in the ARtree (bintree, quadtree, octree).
 * 
 * The key information needed here is 
 *    resolution level
 *    bounds of the original data (array) space represented by this node 
 *       (independent of resolution)
 *    ARNodeBlock - reference to the data at this resolution for this node
 *       the data should behave with delayed binding: the actual data should
 *       not be loaded until it is needed (ideally)
 * 
 * Notes:
 * 1. The resolution level is represented by an integer, which makes sense for
 * wavelet and selection representations, but could get complicated for some
 * other hypothetical data reps. If we ever need those, we would not want to
 * complicate the simpler cases, so we'd probably create a subclass.
 * 
 * 2. bounds: As Sam described it, it can be either the original data bounds 
 * or the "base"bounds. We need to know which it is??? If it can change 
 * depending on whether this is a wavelet or selection representation 
 * (which makes sense), should have either a flag to identify
 * which it is, or (maybe instead or also) 2 bounds variables: 
 * baseBounds, dataBounds.
 * 
 * @author svohr
 */
public class ARNode extends OctreeNode {

    static private int resOffset;

    //private int minBlockSize = pow ( minBlockDimSize + 1, 3 );
    
    /*
     * The resolution independent bounds of the node. Possibly at the
     * original resolution or the base resolution. 
     * 
     */    
    protected ISBounds bounds;
    
    protected ARNodeBlock block;

    // Keep track of resolution levels existing below this one.
    protected int minResolution;
    protected int maxResolution;
    
    public ARNode ()
    {
        bounds = null;
        block = null;
    }
    
    protected ARNode (ISBounds b)
    {
        bounds = b.copy();
        block = null;
        minResolution = 0;
        maxResolution = 0;
    }
    
    protected ARNode (ARNode n)
    {
        for (int i=0; i < children.length; i++)
            this.children[i] = new ARNode ((ARNode)n.children[i]);
        
        bounds = n.bounds.copy();
        block = n.block.copy();
        minResolution = 0;
        maxResolution = 0;
        resOffset = n.resOffset;
    }
    
    public ARNode copy ()
    {
        return new ARNode (this);
    }
    
    public ISBounds getBounds() {
        return bounds;
    }

    public void setBounds(ISBounds b) {
        this.bounds = b.copy();
    }

    public ARNodeBlock getBlock() {
        return block;
    }

    public void setBlock(ARNodeBlock block) {
        this.block = block;
    }
    

    
    private int findMinResolution () 
    {
        if ( this.isLeaf() ) {
            return block.getResolution();
        }
        else {
            int min = block.getResolution();
            for ( ARNode n : (ARNode [])children ) {
                int tmp = n.findMinResolution();
                if ( tmp > min ) {
                    min = tmp;
                }
            }
            return min;
        }
    }
    
    public void dump ()
    {
        this.dump("");
    }
    
    public void dump ( String prefix )
    {
        
        System.err.println( prefix + "AR Node" );
        System.err.println( prefix + "Bounds: " + bounds.toString() );
        
        if ( block != null ) 
            block.dump( prefix );
        else
            System.err.println( prefix + "(no data)" );
    }

    public void dumpAll ( String prefix )
    {
        int childCount = 0;
        
        this.dump(prefix);

        System.err.println( prefix + "Children:");
        
        for ( OctreeNode n : children ) {
            if ( n != null ) {
                n.dumpAll( prefix + "  " );
                childCount++;
            }
        }
        if (childCount == 0)
            System.err.println( prefix + "  (none)" );
    }
    
    public void split ( int res )
    {
        IndexSpaceID low, up;
        
        int minBlockDimSize =  2 * pow(2,res+resOffset); // extra 2 is for base coords
        
        if ( res == 0 ) {
            // this is where the oct tree departs from a true octree:
            // Because there is nothing gained by spliting the region to move 
            // from resolution 1 to resolution 0, we don't and instead add a 
            // child with the same dimensions 
            children[0] = new ARNode (this.bounds);
            
            return;
        }
        
        low = bounds.copyLower();
        up = bounds.copyUpper();
        
        // need 8 bounds for the children
        for ( int i=0; i < children.length; i++ )
            children[i] = new ARNode (this.bounds);
        
        // We use the index of the child to determine if we move its upper
        // or lower coordinate for each dimension. kinda cool!
        for (int i=0; i < 3; i++) {
            
            int size = up.getCoord(i) - low.getCoord(i);

            int diff = ((( size / minBlockDimSize ) / 2 )
                        * minBlockDimSize);
            if ( diff == 0 ) {
                if ( size > minBlockDimSize ) 
                    diff = minBlockDimSize;
                else 
                    diff = size;
            }
            
            for ( int j=0; j < children.length; j++ ) {

                if ( ( j & pow(2,i) ) != 0 ) {

                    ((ARNode)children[j]).getBounds()
                            .setLower( i, low.getCoord(i) + diff );
                    
                }
                if ( ( (children.length - j - 1) & pow(2,i) ) != 0) {

                    ((ARNode)children[j]).getBounds()
                            .setUpper( i, low.getCoord(i) + diff );
                }
            }
        }
        
      // remove nodes that represent just an edge or a face.
        for ( int i=0; i < children.length; i++ ) {
            
            if ( children[i] != null ) {
                ARNode n = ((ARNode) children[i]);
                ISBounds b = n.getBounds();
                for (int d=0; d < b.dim(); d++) {
                    if ( b.getLower(d) >= b.getUpper(d) ) {
                        children[i] = null;
                        break;
                    }
                }
            }
            
        }

    }
    
    public void buildTree ( int res, int offset, 
                            DataCollection [] data, DataCollection [] error )
    {
        ARNode.resOffset = offset;
        
        this.buildTree(res, data, error);
    }


    public void buildTree ( int res, DataCollection [] data, DataCollection [] error )
    {
        int [] test = this.bounds.getUpperArray(); 
                
        if ( res == 0 ) {
            // already at the lowest resolution
            ISBounds b = base2res (this.bounds, res);

            // at original resolution... no error
            ARFloatError e = new ARFloatError (0);
            
            block = new ARNodeBlock ( res, e, b );

            return;
        }
        else {
            // check the contents of the error set
            ISBounds errorRegion = base2err ( this.bounds, res );

            
            if ( ! errorRegion.intersects( error[res].getBounds() ) ) {
                this.buildTree(res-1, data, error);
            }
            else {
                errorRegion.intersectWith( error[res].getBounds() );
            
                // HACK: getting some weird error regions. maybe this will fix it.
                if ( errorRegion.volume() > 0 ) {

                    // build an ARNodeBlock for the region.
                    float maxErrorValue = -1;
                    try {
                        DataBlock d = error[res].subblock(errorRegion);

                        float [] vals = d.getFloats();
                        maxErrorValue = vals[0];

                        // go through the data set, and output 
                        for ( int i=0; i<vals.length; i++ ) {
                            if ( maxErrorValue < vals[i] )
                                maxErrorValue = vals[i];
                        }
                    }
                    catch (Exception e) {
                        System.err.println ( errorRegion.toString() + " " + 
                                             errorRegion.volume() );
                        e.printStackTrace();
                        System.exit(1);
                    }

                    ARFloatError e = new ARFloatError (maxErrorValue);

                    ISBounds b = base2res ( this.bounds, res );
                    b.intersectWith(data[res].getBounds());

                    block = new ARNodeBlock ( res, e, b );

                    // reduce the resolution for the children, if we're at the end
                    // of this resolution.
                    if ( errorRegion.getLower().equals(errorRegion.getUpper()) ) { 
                        res -= 1;
                    }


                    // split this block into subblocks
                    this.split( res );

                    // descend into the children...
                    for ( int i=0; i < children.length; i++ ) {
                        if (children[i] != null) 
                            ((ARNode)children[i]).buildTree (res, data, error);
                    }
                }
                else {
                    this.buildTree(res-1, data, error);
                }
            }

        }
    }
    
    /**
     * Builds a subtree based on a subvolume.
     * @param b the bounds of the subvolume 
     * @return the root of the subtree
     */
    public ARNode subtree (ISBounds b)
    {
        ISBounds vol;
        
        System.err.print(bounds.toString() + " " + b.toString());
        if ( (vol = this.bounds.intersection(b)) != null ) {
            System.err.println(" INTERSECT!");
            ARNode node = this.copy();
            
            
            node.bounds.intersectWith(b);
            
            // get the data subblock
            node.block = this.block.subBlock( 
                            base2res(node.bounds, block.getResolution()) );
                    
            
            // handle the subvolumes
            for (int i=0; i<children.length; i++) {
                if ( this.children[i] != null )
                    node.children[i] = ((ARNode)this.children[i]).subtree(b); 
            }
            
            return node;
        }
        System.err.println(" MISS!");
        
        
        // otherwise return an empty pointer.
        
        return null;
    }

    /**
     * performs a prefix traversal of the tree based on a given error
     * tolerance.
     */
    public void traverse ( ARError e, String prefix )
    {
        this.dump(prefix);

        // if there is a block within the tolerance on this node,
        // do not descend any further. 
        if ( block != null ) {
            if ( block.withinTolerance (e) )
                return;
        }

        for ( int i=0; i < children.length; i++ ) {
            if (children[i] != null)
                ((ARNode)children[i]).traverse(e,prefix + "  ");
        }
    }
    
    /**
     * 
     * @param e
     */
    public ARNode pruneCopy ( ARError e )
    {
        ARNode res = this.copy();
        
        return res.prune(e);
    }
    
    /**
     * 
     * @param e
     */
    public ARNode prune ( ARError e )
    {
      
        // if there is a block within the tolerance on this node,
        // do not descend any further. 
        if ( block != null && block.withinTolerance (e) ) {
            this.clearChildren();
        }
        else {
            
            for ( int i=0; i < children.length; i++ ) {
                if (children[i] != null)
                    children[i] = ((ARNode)children[i]).prune(e);
            }
        }
        return this;
    }
    
    /**
     * NOT DONE YET!
     * @param e
     */
    public void condense ( )
    {
        if ( this.isLeaf() ) {
            minResolution = block.getResolution();
            maxResolution = minResolution;
        }
        else {
            minResolution = Integer.MAX_VALUE;
            maxResolution = Integer.MIN_VALUE;
            
            int missing = 0, nonleafs = 0;
            
            for ( int i=0; i < children.length; i++ ) {
                if (children[i] != null) {
                    ARNode n = (ARNode)children[i];
                    n.condense();
                    if ( n.minResolution < this.minResolution ) 
                        this.minResolution = n.minResolution;
                    if ( n.maxResolution > this.maxResolution )
                        this.maxResolution = n.maxResolution;
                    if ( !n.isLeaf() )
                        nonleafs += 1;
                }
                else {
                    missing += 1;
                }
            }
            if ( maxResolution == minResolution ) {
                if ( missing == 0 && nonleafs == 0 ) {
                    // Consensus on resolution. Merge the children.
                    ARError err = ((ARNode)children[0]).getBlock().getError();
                    for ( int i=0; i < children.length; i++ ) {
                        ARNode n = (ARNode)children[i];
                        if (err.withinTolerance( n.getBlock().getError() ) )
                            err = n.getBlock().getError();
                    }

                    ARNodeBlock lb = ((ARNode)children[0]).getBlock();
                    ARNodeBlock ub = ((ARNode)children[7]).getBlock();

                    ISBounds b = new ISBounds ( lb.getBounds().getLower(), 
                                                ub.getBounds().getUpper() );

                    this.setBlock(new ARNodeBlock (minResolution, err, b));

                    this.clearChildren();
                }
                else if ( children[0] != null ) {
                     ARNode n = (ARNode)children[0];
                     if ( n.getBounds().equals(bounds) ) {
                         this.setBlock(n.getBlock());
                         this.clearChildren();
                     }
                }
            }
        }
    }
    
    public void getLeafBlocks ( List<ARNodeBlock> l ) {
        if ( this.isLeaf() ) {
            l.add(block);
        }
        else {
            for ( int i=0; i < children.length; i++ ) {
                if ( children[i] != null )
                    ((ARNode)children[i]).getLeafBlocks(l);
            }
        }
    }
    
        
    public void getLeafBlocks ( List<ARNodeBlock> l, ARError e ) {
        if ( block != null && block.withinTolerance (e) ) {
            l.add(block);
        }
        else {
            for ( int i=0; i < children.length; i++ ) {
                if (children[i] != null)
                    ((ARNode)children[i]).getLeafBlocks(l, e);
            }
        }
    }
    
    /**
     * Converts a bounds in base coordinates to a bounds in a data set
     * at the given resolution level (n).
     */
    private ISBounds base2res ( ISBounds b, int n )
    {
        ISBounds ret = b.copy();

        for ( int d=0; d < 3; d++ ) { // for each dimension
            
            ret.setLower ( d, b.getLower (d) / ( pow(2,n+1) ));
            ret.setUpper ( d, b.getUpper (d) / ( pow(2,n+1) ));
        }
        //System.err.println ("d " + b.toString() + " >> " + ret.toString() );

        return ret;
    }

    /**
     * Converts a bounds in base coordinates to a bounds in an error set
     * for the data at the given resolution level (n).
     */
    private ISBounds base2err ( ISBounds b, int n )
    {
        ISBounds ret = b.copy();

        for ( int d=0; d < 3; d++ ) { // for each dimension
            
            ret.setLower ( d, b.getLower (d) / ( pow(2,n+1+resOffset) ));
            ret.setUpper ( d, (b.getUpper (d)-1) / ( pow(2,n+1+resOffset) ));
            // Note: error regions are inclusive along the lower edges and 
            // exclusive along the upper. Subtracting 1 should ensure that
            // there is no overlap.
        }
        
        //System.err.println ("e " + b.toString() + " >> " + ret.toString() );

        return ret;
    }

    private int pow (int n, int e) 
    {
        int result = 1;
        for ( int i=0; i < e; i++ ) 
            result *= n;

        return result;
    }

    public static void main( String[] args )
    {
        ARNode root;
        
        int [] low = {0,0,0};
        int [] up = {14,14,14};
        
        DataSource [] data = new DataSource [3];
        DataSource [] error = new DataSource [3];
        
        data[0] = DataSource.create("data_0","data_0.fdl");
        data[1] = DataSource.create("data_1","data_1.fdl");
        data[2] = DataSource.create("data_2","data_2.fdl");
        
        for ( DataSource d : data)
            d.activate();
        
        error[1] = DataSource.create("error_1_2","error_1_2.fdl");
        error[2] = DataSource.create("error_2_3","error_2_3.fdl");

        for ( int i=1; i < error.length; i++ )
            error[i].activate();
        
        ISBounds b = new ISBounds (low, up);
        
        root = new ARNode (b);
        
        root.buildTree(2, data, error);
        root.dumpAll("");
        
        //int [] l = {8,8,8};
        //int [] u = {10,10,10};
        //ARNode s = root.subtree(new ISBounds (l,u));
        //root.dumpAll("");
        //s.dumpAll("");
        //s.traverse(new ARFloatError (7), "");
        
        return;
    }
}


