/*
 * Class for writing AR xml files based on XFDLWriter class.
 * 
 * 
 * Edit history (beginning 12/22/11)
 * 02/09/12 rdb: replaced all the empty catch (IOException blocks) with 
 *               code that reports the exception and continues.
 * 
 */
package ar;

import edu.unh.sdb.datasource.ISBounds;
import edu.unh.sdb.common.*;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Enumeration;
import java.util.LinkedList;
import java.util.Properties;

/**
 * Writes AR xml files. Could be replaced with a better class.
 *
 * 
 * @author Sam Vohr
 */
public class ARWriter 
{
   File outputFile = null;
   PrintWriter outputWriter = null;
   LinkedList<String> tagList = new LinkedList<String>();
   private String tabs;

   public ARWriter() 
   {
      super();
      tabs = "";
   }

   public File getOutputFile() 
   {
      return outputFile;
   }

   public String getOutputFilename() 
   {
      if (outputFile == null) {
         return null;
      } else {
         return outputFile.getAbsolutePath();
      }
   }

   public void setOutputFile(File out) throws IOException 
   {
      if (out == null) {
         return;
      }
      if ( out.isDirectory() ) {
         throw new IOException(out.getAbsolutePath() + " is a directory." );
      }
      if (outputFile != null) {
         this.close();
      }
      outputFile = out;
   }

   public void setOutputFile(String out) throws IOException 
   {
      File outF = new File( out );
      //rdb: 12/2011: added warning if file exists, should be SDBError
      if ( outF.isFile() ) {
         new SDBError( SDBError.Warning, "Overwriting file: " + out );
      }
      setOutputFile(outF);
   }

   public void close() 
   {
      try {
         if (tagList.size() != 0)
            closeAllTags();
         if (outputWriter != null)
            outputWriter.close();
      } catch (IOException e)  { reportIOX( e, "close()" ); }  
      finally {
         outputWriter = null;
         outputFile = null;
      }
   }

   private void writeHeader() throws IOException 
   {
      spew("<?xml version=\"1.0\" encoding=\"UTF-8\"?>", true);
      spew("<!DOCTYPE ar_descriptor PUBLIC \"-//SDB//DTD//EN\"  \"ardl.dtd\">", true);
   }

   private void pushTag(String tagName) 
   {
      tagList.add(tagName);
      tabs = tabs + "    ";
   }

   private String popTag() 
   {
      tabs = tabs.substring(4);
      return (String) tagList.removeLast();
   }

   private void spew(String text, boolean newLine) throws IOException 
   {
      if (outputWriter == null) {
         if (outputFile != null) {
            outputWriter =
                    new PrintWriter(new FileOutputStream(outputFile));
         } else {
            outputWriter = new PrintWriter(System.out);
         }
      }
      if (!newLine) {
         outputWriter.print(text);
      } else {
         outputWriter.println(text);
      }
   }

   /**
    * Writes a tag to the file, given the arguments.
    * 
    * @param tagName
    * @param tagAttrib
    * @param tagData
    * @param isEmpty
    * @throws java.io.IOException
    */
   private void writeTag(String tagName, Properties tagAttrib,
           String tagData, boolean isEmpty)
           throws IOException 
   {
      // check if everything has been set up already...
      if (tagName == null) {
         return;
      }
      //if (outputFile == null)
      //    return;
      if (outputWriter == null) {
         writeHeader();
      }

      // print tag name
      spew(tabs + "<" + tagName, false);

      // print all the tag's attributes
      if (tagAttrib != null) {
         for (Enumeration e = tagAttrib.propertyNames(); e.hasMoreElements();) {
            String keyName = e.nextElement().toString();
            String keyValue = tagAttrib.getProperty(keyName);
            spew(" " + keyName + "=\"" + keyValue + "\"", false);
         }
      }

      if (!isEmpty) {
         spew(">", true);             // end the start tag

         if (tagData != null) {
            spew(tagData, true);    // print data if available
         }
         pushTag(tagName);           // push to be closed later
      } else {
         // close the tag
         spew("/>", true);
      }
   }

   /**
    * Closes the most recent tag.
    * @throws java.io.IOException
    */
   private void closeLastTag() throws IOException 
   {
      String tag = popTag();

      spew(tabs + "</" + tag + ">", true);
   }

   /**
    * closes all open tags.
    * @throws java.io.IOException
    */
   private void closeAllTags() throws IOException 
   {
      while (tagList.size() <= 0) {
         closeLastTag();
      }
   }

   /**
    * Procedure for adding a property with string value.
    * @param props
    * @param keyName
    * @param keyValue
    */
   private void setProp(Properties props, String keyName, String keyValue) 
   {
      if (keyValue != null || keyValue.trim().length() <= 0) {
         props.setProperty(keyName, keyValue);
      }
   }

   /**
    * Procedure for adding a property with an int value.
    * @param props
    * @param keyName
    * @param keyValue
    */
   private void setProp(Properties props, String keyName, long keyValue) 
   {
      props.setProperty(keyName, String.valueOf(keyValue));
   }

   /**
    * Procedure for adding a property with a float value.
    * @param props
    * @param keyName
    * @param keyValue
    */
   private void setProp(Properties props, String keyName, float keyValue) 
   {
      props.setProperty(keyName, String.valueOf(keyValue));
   }

   private void writeBounds(ISBounds b) 
   {
      Properties boundsProps = new Properties();
      int[] coords = b.getLowerArray();
      setProp(boundsProps, "lower",
              new String("" + coords[0] + "," + coords[1] + "," + coords[2]));
      coords = b.getUpperArray();
      setProp(boundsProps, "upper",
              new String("" + coords[0] + "," + coords[1] + "," + coords[2]));
      try {
         this.writeTag("bounds", boundsProps, null, true);
      } catch (IOException e)  { reportIOX(e, "bounds tag" ); } 
   }

   private void writeError(ARError e) 
   {
      Properties errorProps = new Properties();
      setProp(errorProps, "type", e.typeString());
      setProp(errorProps, "value", e.valueString());
      try {
         this.writeTag("error", errorProps, null, true);
      } catch (IOException ioex)  { reportIOX( ioex, "error tag" ); } 
   }

   private void writeBlock(ARNodeBlock b) 
   {
      Properties blockProps = new Properties();

      // resolution (for reference back to data files)
      setProp(blockProps, "resolution", b.getResolution());
      try {
         writeTag("block", blockProps, null, false);
      } catch (IOException e)  { reportIOX( e, "block tag" ); } 

      // dimensions (in index space coords)
      writeBounds(b.getBounds());

      // error
      writeError(b.getError());

      // close the tag
      try {
         this.closeLastTag();
      } catch (IOException e)  { reportIOX( e, "closeLastTag" ); } 
   }

   private void writeNode(ARNode n) 
   {
      if (n == null) {
         return;
      }

      Properties nodeProps = new Properties();

      try {
         writeTag("node", nodeProps, null, false);
      } catch (IOException e)  { reportIOX( e, "node tag" ); } 

      // dimensions (in base coords)
      writeBounds(n.getBounds());

      // write the tags for the block(s) it contains. (right now just 1)
      writeBlock(n.getBlock());

      // write the tags for the child nodes
      OctreeNode[] children = n.getChildren();
      for (int i = 0; i < children.length; i++) {
         writeNode((ARNode) (children[i]));
      }


      // close the tag
      try {
         this.closeLastTag();
      } catch (IOException e)  { reportIOX( e, "closeLastTag" ); } 
   }

   public void writeAR( ARDataSource d ) 
   {
      // write ARDescriptor tag
      Properties arProps = new Properties();
      setProp(arProps, "resolutions", d.getResolutions());
      try {
         writeTag("ar_descriptor", arProps, null, false);
      } catch (IOException e) { reportIOX( e, "ar_descriptor tag" ); }

      // write references to existing multires files
      String[] sources = d.getDataFDLNames();
      for (int i = 0; i < sources.length; i++) {
         Properties sourceProps = new Properties();
         setProp(sourceProps, "desc", sources[i]);
         setProp(sourceProps, "resolution", i);
         try {
            writeTag("source", sourceProps, null, true);
         } catch (IOException e) { reportIOX( e, "source tag" ); }
      }

      // write tree tags

      try {
         writeTag("tree", null, null, false);
      } catch (IOException e) {  }

      writeBounds(d.getBounds());
      writeNode(d.getRoot());

      // close the last of the tags
      try {
         this.closeLastTag();
         this.closeLastTag();
      } catch (IOException e) { reportIOX( e, "closeLastTag" ); }

      // all done
      return;
   }
   private void reportIOX( IOException e, String note )
   {
      System.err.println( "ARWriter ignoring IOException related to: " + note 
                          + " ---> " + e.getMessage() );
   }
}
