/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

/**
 * SparseDataSink - interface to the file system for creating a new 
 *                   SparseDataSource
 * A SparseDataSource only stores a subset of the full datasource; it is 
 * intended to be used as part of an AR representation, but in principle, it
 * could also be used for storing data. The current implementation partitions
 * the full geometry of the conceptual datasource into equal size blocks 
 * (except for the boundaries (maybe). A block in the partition is either 
 * completely written or completely absent in the Sparse version. A mapping 
 * table maps virtual block indexes to physical block indexes in the sparse file.
 * "Partial" blocks are stored as full size blocks, with the "real" data in
 * the block stored at the start of the linear array used to store the block.
 * i.e. if a full block is 4x4x4 datums, but a partial (edge) block is 2x2x1, 
 * the "real" 4 datums will be stored in the first 4 entries of the 1D view 
 * of the 4x4x4 block that is stored on the disk. We do this in order to
 * keep the retrieval of data efficient since the mapping of a physical 
 * index to a file seek address is always physicalBlockNumber * blockSize.
 * 
 * The mapping data can be added incrementally or all at once.
 * 
 * @author rdb, derived from Rhodes' DataSink class
 * Created 01/04/12
 */
package ar;

import edu.unh.sdb.common.*;
import edu.unh.sdb.datasource.*;

import java.io.*;
import java.util.*;

public class SparseDataSink extends DataSink
{
   //----------------- instance variables --------------------------
   protected HashMap<Integer,Integer> blockMapper = null;
   protected int[] blockSize     = null;
   protected int   blockDatumCount = 0;  // size of block in # Datums
   protected int   blockByteSize   = 0;   // size if each block in bytes
   /* 
    * ---------- inherited instance variables from DataSink --------------
	protected RandomAccessFile raFile;
	protected String filename;
	protected long [] axisWeights;
	protected RecordDescriptor recordDescriptor;
	protected AxisOrdering ordering;
	protected ISBounds bounds;
	protected int dimensionality;
	protected int numFields;
	protected int datumSize;
  */

   
   //----------------- constructors --------------------------------
   /**
    * Create a DataSink to serve as the interface for writing a new Datasource
    * to a file.
    * 
    * @param filename
    * @param RecordDescriptor rd       - 
    * @param ISBounds bounds - bounds of complete (virtual) d.s.
    * @param int[] blockSize - sizes (in each dim) of blocks
    */
   public SparseDataSink( String filename, RecordDescriptor rd, 
                          ISBounds bounds, int[] blkSize )
   {
      super( filename, rd, bounds );
      blockMapper = new HashMap<Integer,Integer>();   
      blockSize = blkSize;   // should this be a copy?
   }

	/**
    * Not used yet. Is it necessary?
    * 
    * @param filename
    * @param rd 
    */
	protected SparseDataSink( String filename, RecordDescriptor rd )
   {
      super( filename, rd );
      blockMapper = new HashMap<Integer,Integer>();     
	}	
	
   /**
    * Define a mapping from a virtual block number to a physical block number
    * 
    * @param int vBlock - block in the original (virtual) file
    * @param int pBlock - block location in the sparse file
    */
   protected void addMap( int vBlock, int pBlock )
   {
      blockMapper.put( vBlock, pBlock );
   }
   protected void setMapper( HashMap<Integer,Integer> newMap )
   {
      blockMapper = newMap;
   }
   /**
    * Write a Datum object to its location in the sparse file. It is an 
    * error If the location of the Datum is NOT defined in the SparseDataSource 
    * mapper,  We probably should throw an exception, but for now we just
    * generate an SDB error and return.
    * 
    * 
    * @param Datum d 
    * @param IndexSpaceID id 
    */
	public void write( Datum d, IndexSpaceID id )
   {	
      int datumOffset = this.computeOffset( id );
      int vBlockId    = datumOffset / blockDatumCount;
      Integer pBlockInt = blockMapper.get( vBlockId );
      
      if ( pBlockInt == null )  // bad reference
      {
         SDBError err = new SDBError( SDBError.Recoverable, 
                 "Data location " + id + " is not in sparse datasource ");
      }
      else
      {
         int pBlockId = pBlockInt.intValue();
         int vBlockOffset = vBlockId * blockByteSize;
         int pBlockOffset = pBlockId * blockByteSize;
         int vSeekOffset = datumOffset * this.datumSize;
         int pSeekOffset = ( vSeekOffset - vBlockOffset ) + pBlockOffset;
         writeDatum( d, pSeekOffset );
      }
   }
   
   protected void writeDatum( Datum d, int seekOffset )
   {
      if ( recordDescriptor.getNumUsedTypes() > 1 )
         writeDatumMixed( d, seekOffset );
      else
         writeDatumUniform( d, seekOffset );
   }
   
   /**
    * Write a Datum object to its correct position in the file; this version
    * works for mixed type Datum objects.
    * 
    * Note: this method should be moved into the parent class
    *       it could also be made much more efficient by checking if all
    *       values in Datum are of same type and doing no conversion.
    * 
    * @param d
    * @param offset 
    */
	protected void writeDatumMixed( Datum d, int seekOffset )	
   {
		double value;
		int storageType;
		
		try {
			this.raFile.seek( seekOffset);
			
			for( int i=0; i < this.numFields; i++ )
         {			
				storageType = recordDescriptor.getAttribute(i).getStorageType();
				
				value = d.getDouble(i);
				
				switch(storageType){
				
					case StorageType.Float:
						this.raFile.writeFloat( (float) value);
						break;
						
					case StorageType.Double:
						this.raFile.writeDouble(value);
						break;
						
					case StorageType.Int:
						this.raFile.writeInt((int)value);
						break;
						
					case StorageType.Short:
						this.raFile.writeShort((short)value);
						break;
						
					case StorageType.Byte:  // or should this be unsigned byte?
						this.raFile.writeShort((byte)value);
						break;
						
					default:
						new InternalFaultError(SDBError.Fatal,
                    "Unsupported storage type code in DataSink.write(Datum, IndexSpaceID):"+storageType);
						break;
				}
			}
			
			
		} catch (IOException e){
			new FileError(SDBError.Recoverable, "IO Error in write(): "+e.getMessage());
		}
	}

   /**
    * Write a Datum object to its correct position in the file; this version
    * works only for uniform Datum objects -- and is more efficient in this case
    * 
    * Note: this method should be moved into the parent class
    *       it could also be made much more efficient by checking if all
    *       values in Datum are of same type and doing no conversion.
    * 
    * @param d
    * @param offset 
    */
	protected void writeDatumUniform( Datum d, int seekOffset )	
   {
      int storageType;
		
      try {
         this.raFile.seek( seekOffset );
         storageType = recordDescriptor.getAttribute(0).getStorageType();

         switch (storageType) {
            case StorageType.Float:
               float[] fvals = d.getFloats();
               for (int i = 0; i < fvals.length; i++)
                  this.raFile.writeFloat(fvals[ i]);
               break;

            case StorageType.Double:
               double[] dvals = d.getDoubles();
               for (int i = 0; i < dvals.length; i++) 
                  this.raFile.writeDouble(dvals[ i]);
               break;

            case StorageType.Int:
               int[] ivals = d.getInts();
               for (int i = 0; i < ivals.length; i++) 
                  this.raFile.writeInt(ivals[ i]);
               break;

            case StorageType.Short:
               short[] svals = d.getShorts();
               for (int i = 0; i < svals.length; i++) 
                  this.raFile.writeShort(svals[ i]);
               break;

            case StorageType.Byte:  // or should this be unsigned byte?
               byte[] bvals = d.getBytes();
               this.raFile.write(bvals);
               break;

            default:
               new InternalFaultError(SDBError.Fatal,
                       "Unsupported storage type code in DataSink.write(Datum, IndexSpaceID):" + storageType);
               break;
         }
      } catch (IOException e) {
         new FileError(SDBError.Recoverable, "IO Error in write(): " + e.getMessage());
      }
   }


   /**
    * return the virtual block index for the virtual block containing the 
    * IndexSpaceID
    * 
    * @param IndexSpaceID isid
    * @return 
    */
   protected int vBlockNumber( IndexSpaceID isid )
   {
      return computeOffset( isid );
   }
   /**
    * return the virtual block index for the virtual block containing the space 
    * position defined by the array
    * 
    * @param int[] coordArray
    * @return 
    */
   protected int vBlockNumber( int[] coordArray )
   {
      return computeOffset( coordArray ) / blockDatumCount;
   }
    
    //---------------------------- main ---------------------------------
    /*
     * unit test
     */
    public void main( String[] args )
    {
       RecordDescriptor rd = null;
       SparseDataSink test = new SparseDataSink( "test", rd );
    }
}