/*
 * SparseDataTest: test the SparseDataSource class
 */

package ar;

/**
 * SparseDataTest - Tests datasources created as SparseDataSources 
 *    command line arguments:
 *       origDS       - xfdl for the original datasource
 *       sparseDS     - xfdl for the sparse version of origDS
 *       bounds       - optional subblock to read
 * 
 * The code 
 *    1. if bounds not specified, generates random subblock requests
 *    2. if the subblock is entirely represented in the sparse data source, 
 *       it reads the same from both datasources and verifies that the
 *       data is the same.
 * 
 * @author rdb
 * created 01 Feb 2012
 * 
 */

import java.io.*;
import java.nio.*;
import java.util.*;
import edu.unh.sdb.common.*;
import edu.unh.sdb.datasource.*;

public class SparseDataTest 
{
   //------------- class variables ------------------------------
   // Command line args
   static String  origDSname = null;
   static String  sparseDSname = null;
   static int[]   locArgs = null;   
   static int     nRandom = 0;
   static boolean datumTest = false;   // default is block test
   static boolean userInput = false;   // batch input is default
   
   // processing variables
   static DataSource       origDS;
   static DataSource       sparseDS; 
   static ISBounds         firstBounds = null;
   static IndexSpaceID     firstLoc = null;
   static HashMap<Integer, Integer> blkMap = new HashMap<Integer, Integer>();
   static ArrayList<IndexSpaceID> blkStart = new ArrayList<IndexSpaceID>();
   static int    inBlkNum;  // # blocks in full file
   static Scanner          userIn = null;
   
   static byte[]           zeroData = null;
   
   //---------------------- usage()  --------------------------
   static void usage()
   {
      String[] msg = 
      {
         "Usage: java SparseDataTest fullDSxfdl sparseDSxfdl [-d] [-i] [-r n] [s1 ... sk]",
         "    fullDSxfdl - xfdl file for original full ds",
         "    sparseDSxfdl - xfdl file for sparse version of fullDS",
         "    -d - test datum, rather than subblock",
         "    -i - allow user to interactvely enter location or block specs",
         "    s1 ... sk - defines data to to access.",
         "              if datum test, k = d and s1..sd is a position",
         "              if block test, k = 2d and s1..s2d is a bounds",
         "    -r n - generate n random datum or subblock specifications to test",
         "    ",
         "Opens both data sources, extracts datums/subblock(s) from them both and",
         "  compares the data in them -- where it is shared. If no bounds",
         "  specified on command line and no -r switch, default is -r 1"
      };
      for ( int i = 0; i < msg.length; i++ )
         System.out.println( msg[ i] );
      System.exit( 0 );
   }
   //---------------------- main -----------------------------------   
   public static void main( String[] args )
   {
      readArgs( args );
      
      // open the input datasource
      origDS = DataSource.create( "fullDS", origDSname );
      if ( origDS == null )
      {
         System.err.println( "Unable to create a data source from " + origDSname );
         System.exit( 4 );
      }
      origDS.activate();
      System.err.println( "Creating sparse: " );
      sparseDS = DataSource.create( "sparseDS", sparseDSname );
      if ( sparseDS == null )
      {
         System.err.println( "Unable to create a data source from " + sparseDSname );
         System.exit( 4 );
      }
      else if ( !( sparseDS instanceof SparseDataSource ))
      {
         System.err.println( "Sparse ds name is not a sparse datasource. "
                             + "Continuing anyway;  maybe that's useful too." );
      }
      sparseDS.activate();
      
      // check the bounds
      if ( locArgs != null && !datumTest )
      {
         if ( origDS.dim() == 2 * locArgs.length )
            firstBounds = new ISBounds( locArgs );
         else
            System.err.println( "Bounds spec does not have correct # entries. " +
                                " It needs 2 times the dimension of the data." +
                                " Ignoring." );
      }
      else if ( locArgs != null && datumTest )
      {
         if ( origDS.dim() == locArgs.length )
            firstLoc = new IntegerIndexSpaceID( locArgs );
         else
            System.err.println( "Location spec does not have correct # entries. " +
                                " It needs same as the dimension of the data." +
                                " Ignoring." );
      }
      if ( userInput )
         userIn = new Scanner( System.in );
      if ( datumTest )
         testDatums();
      else
         testBlocks();
   } 
   //----------------- testDatums ------------------------------
   private static void testDatums()
   {
      if ( firstLoc != null )
         testDatum( firstLoc );
     
      for ( int r = 0; r < nRandom; r++ )
      {
         IndexSpaceID loc = makeRandomLoc();
         if ( loc != null )
            testDatum( loc );
      }
      if ( userInput )
      {
         int[] pos = userArgs( "Datum position", origDS.dim() );
         while ( pos != null )
         {
            testDatum( new IntegerIndexSpaceID( pos ));
            pos = userArgs( "Datum position", origDS.dim() );
         }
      }
   }
   
   //----------------- testBlocks -------------------------------
   private static void testBlocks()
   {
      if ( firstBounds != null )
         testBlock( firstBounds );
     
      for ( int r = 0; r < nRandom; r++ )
      {
         ISBounds bounds = makeRandomBounds();
         if ( bounds != null )
            testBlock( bounds );
      }
      if ( userInput )
      {
         int[] pos = userArgs( "subblock bounds", 2 * origDS.dim() );
         while ( pos != null )
         {
            int[] low = { pos[0], pos[1], pos[2] };
            int[] up  = { pos[3], pos[4], pos[5] };
            testBlock( new ISBounds( low, up ) );
            pos = userArgs( "subblock bounds", 2 * origDS.dim() );
         }
      }
   }
   //----------------- userArgs( String, int ) -------------------------------
   private static int[] userArgs( String prompt, int num )
   {
      int[] data = new int[ num ];
      System.out.print( prompt + ": enter " + num + " ints> " );
      int n = 0;
      String line = userIn.nextLine();
      
      if ( line != null && line.length() > 0 )
      {   
         Scanner lineIn = new Scanner( line );
         while ( lineIn.hasNextInt() && n < num )
         {
            data[ n++ ] = lineIn.nextInt();
         } 
      }
      if ( n == 0 )
         return null;
      if ( n < num )
         System.err.println( "\n*** Too few values entered" );
      System.out.println( "-----" );
      return data;
   }
   //----------------- makeRandomBounds() -------------------------------------
   static ISBounds makeRandomBounds()
   {
      return null;
   }
   //----------------- makeRandomLoc() -------------------------------------
   static IndexSpaceID makeRandomLoc()
   {
      return null;
   }
   //----------------- testBlock( ISBounds ) -------------------------
   /**
    * Read the block from both datasources and compare if results are the same
    * 
    * @param ISBounds
    */
   static void testBlock( ISBounds isb )
   {
      DataBlock origDB   = origDS.subblock( isb );     
      DataBlock sparseDB = sparseDS.subblock( isb ); 
      int badMatches = compareBlocks( origDB, sparseDB );
      if ( badMatches == 0 )
         System.out.println( "Match for " + isb );
      else
      {
         System.out.println( badMatches + " values don't match ");
         System.out.println( "--- original :\n" + origDB );
         System.out.println( "--- sparse   :\n" + sparseDB );
      }
   }
   
   //----------------- testDatum( IndexSpaceID ) -------------------------
   /**
    * Read the block from both datasources and compare if results are the same
    * 
    * @param ISBounds
    */
   static void testDatum( IndexSpaceID loc )
   {
      Datum origDatum   = origDS.datum( loc );  
      Datum sparseDatum = sparseDS.datum( loc );
      if ( origDatum.equals(  sparseDatum ) )
         System.out.println( "Match at " + loc );
      else
         System.out.println( origDatum + " --> " + sparseDatum );
   }
   //--------------- compareBlocks( DataBlock, DataBlock ) -----
   private static int compareBlocks( DataBlock one, DataBlock two )
   {
      int misMatch = 0;
      float[] f1 = one.getFloats();
      float[] f2 = two.getFloats();
      for ( int i = 0; i < Math.min( f1.length, f2.length ); i++ )
      {
         if ( f1[ i ] != f2[ i ] )
            misMatch++;
      }
      misMatch += Math.abs( f1.length - f2.length );
      return misMatch;
   }
   //--------------- readArgs -----------------------------------
   static void readArgs( String[] args )
   {
      // check for valid argument count
      if ( args.length < 2 )
         usage();
      
      boolean stillParse = true;
      int iarg = 0;
      for ( iarg = 0; iarg < args.length && stillParse; iarg++ )
      {
         String arg = args[ iarg ];
         if ( arg.equals( "-d" ))
            datumTest = true;
         else if ( arg.equals( "-i" ))
            userInput = true;
         else if ( arg.equals( "-r" ))
         {
            if ( iarg < args.length - 1 )
               nRandom = getArg( args, ++iarg, 1 );
            else
            {
               System.err.println( "No value for -r switch; assuming 10" );
               nRandom = 1;
            }
         }
         else if ( origDSname == null )
            origDSname = arg;
         else if ( sparseDSname == null )
            sparseDSname = arg;
         else
            stillParse = false; // handled required args and -r switch
      }
      
      if ( iarg < args.length )  // check for bounds args
         locArgs = getArgArray( args, iarg, args.length - 1, -1 );
      if ( locArgs != null && locArgs.length == 0 )
         locArgs = null;
      
      // if no bounds and no random count specified, make random count 1
      if ( locArgs == null && nRandom == 0 )
         nRandom = 1;
   }
   //---------------------- getArg( String[], int, int ) -----------------
   /**
    * This is a utility method for accessing one of the command line
    * string arguments and converting it to an int. It accepts the 
    * entire array of arguments in String form, an integer indicating 
    * which is to be converted to an int and a default value to be 
    * returned if this argument was not on the command line, or if 
    * the specified String is not a valid integer representation.
    */
   public static int getArg( String[] args, int which, int defaultVal )
   {
      try
      {
         // The Integer class has a class method that will convert a String
         //   to an int -- if it is a valid representation of an  integer. 
         return Integer.parseInt( args[ which] );
      }
      catch ( ArrayIndexOutOfBoundsException oob )
      {
         // If there is no args[ which ] element, Java "throws" an exception.
         // This code "catches" the exception and handles it gracefull.
         // In this case, it is not an error. The parameter is optional
         // and there is a specified default value that is returned.
      }
      catch ( NumberFormatException nfe )
      {
         // If the string is not a valid representation of an integer
         // (such as 4Qd3), Integer.parseInt throws a "NumberFormatException".
         // Again, this code catches the exception, gives an error message
         // and uses the default value.

         System.err.println( "Error: improper command line argument " +
                              which + " = " + args[ which] +
                             
                             ".  It should be an integer; using default value: " +
                              defaultVal );
      }
      return defaultVal;
   }
   //----------------- getArgArray( String[], int, int, int ) --------------

   /**
    * This is a utility method for converting multiple command line
    * string arguments to an array of ints. It accepts the 
    * entire array of arguments in String form, integers indicating the
    * range of arguments to process and returns an array of ints for the
    * arguments in that range. If any argument is missing or is an invalid
    * integer, error messages will be generated and the returned value will
    * be set to the last argument (defVal)
    */
   public static int[] getArgArray( String[] args, int start, int end,
                       int defVal )
   {
      int[] vals = new int[ end - start + 1 ];
      int argIndex = start;
      for ( int i = 0; i < vals.length; i++ )
      {
         vals[ i] = getArg( args, argIndex, defVal );
      }
      return vals;
   }
}

