/*
 * SparseDataTestGen -- Utility to generate a SparseDataSource from an
 *        existing DataSource. Is a template for a AR-based tool
 *        that would use ARDataSource information.
 * 
 * History
 * 02/12/12 rdb: Changed syntax of xfdl information: replaced "ChunkSize"
 *               attribute with "BlockSize" to allow chunking on top of 
 *               blocking.
 * 02/09/12 rdb: Changed mapping information to include possibly optional
 *               block size information in the mapping entry
 * 12/5-22/2011 rdb created
 */

package ar;

/**
 * SparseDataTestGen - generates data for testing SparseDataSink 
 *    command line arguments:
 *       inDS         - xfdl for existing datasource to sparsify
 *       outName      - prefix to be used for output ds files
 *       perCentKeep  - integer (1-100) % of data blocks to keep
 *       s1 s2 ... sk - integers representing size of blocks
 *                      if k < dim of data source, the last specified size
 *                      is used for all remaining dimensions.
 * 
 * This class defines the format of the fdl information stored in the xfdl
 * file and FDLDescriptor object:
 *     
 * String format:  
 *   k@(s,r,c)=(s0,s1,s2)->n
 * 
 * where n is the block index in the sparse data set
 *       (s,r,c) (lower coord) is the "origin" of the block in original space
 *       (s0,s1,s2) sizes of block (this might become optional, so make it so
 *       n is the block index in the original space
 *   and the block size is optional and the = can be there even if the block
 *   size is not. However, this program either includes the = AND the size or
 *   neither.
 *
 * 
 * WARNINGS and LIMITATIONS
 *   1. This code assumes that the input datasource is not chunked.
 *   2. Although it should work with any other datasource, it relies heavily
 *      on the Values family of classes; it's not clear how robust that code
 *      is and this code has only been tested on simple datasources. So,
 *      for example, I haven't tested it with datasources stored by attribute,
 *      CompositeDataSources and datasources with MixedType datums.
 *   3. This implementation adds a separate fdl field for block specifciation
 *      from the one used for chunking. This would allow us later to do chunking
 *      as well as the blocking here; that probably only makes sense if 
 *      ChunkedDataSource is re-implemented as a wrapper around any other
 *      DataSource -- whic is probably a good idea, but I doubt we'll ever
 *      get to that.
 * 
 * @author rdb
 * created 6-10 Jan 2012
 * 
 */

import java.io.*;
import java.nio.*;
import java.util.*;
import edu.unh.sdb.common.*;
import edu.unh.sdb.datasource.*;

public class SparseDataTestGen 
{
   //------------- class variables ------------------------------
   // Command line args
   static String inDSname = null;
   static String outPrefix = null;   
   static int    percentKeep = 0;
   static int[]  sizeArgs;
   
   // processing variables
   static int[]  blockSize;   
   static DataSource inds;
   static HashMap<Integer, Integer> blkMap = new HashMap<Integer, Integer>();
   static ArrayList<IndexSpaceID> blkStart = new ArrayList<IndexSpaceID>();
   static ArrayList<IndexSpaceID> blkSizes = new ArrayList<IndexSpaceID>();
   static int    inBlkNum;  // # blocks in full file
   
   static RandomAccessFile outData;
   static byte[]           zeroData = null;
   
   //---------------------- usage()  --------------------------
   static void usage()
   {
      String[] msg = 
      {
         "Usage: java SparseDataTestGen inxfdl outPrefix percent s1 ... sk",
         "    inxfdl - fdl file for existing datasource to sparsify",
         "    outPrefix - prefix for output files (xfdl and bin)",
         "    percent - int defining % of blocks to keep (randomly)",
         "              100 will essentially chunk the file with padding",
         "    s1 ... sk - size of data blocks. if k < ds dimension",
         "              last specified size is used for missing ones."
      };
      for ( int i = 0; i < msg.length; i++ )
         System.out.println( msg[ i] );
      System.exit( 0 );
   }
   //---------------------- main -----------------------------------   
   public static void main( String[] args )
   {
      readArgs( args );
      
      // open the input datasource
      inds = DataSource.create( "fullDS", inDSname );
      if ( inds == null )
      {
         System.err.println( "Unable to create a data source from " + inDSname );
         System.exit( 4 );
      }
      inds.activate();
      
      // check the bounds
      if ( inds.dim() == sizeArgs.length )
         blockSize = sizeArgs;
      else 
      {
         if ( inds.dim() < sizeArgs.length )
            System.err.println( "Too many size arguments; extras ignored." );
         blockSize = new int[ inds.dim() ];
         for ( int i = 0; i < blockSize.length; i++ )
         {
            if ( i < sizeArgs.length )
               blockSize[ i ] = sizeArgs[ i ];
            else
               blockSize[ i ] = sizeArgs[ sizeArgs.length - 1 ];
         }
      }
      // open output file
      openOut();
      
      // generate the sparse blocked data file
      makeSparseData();
      
      // generate the xfdl file for the sparse ds
      makeSparseXfdl();
      
   }  
   //--------------- readArgs -----------------------------------
   static void readArgs( String[] args )
   {
      // check for valid argument count
      if ( args.length < 4 )
         usage();
      
      // save arg 0 as input file; it will be checked by open attempt
      inDSname = args[ 0 ]; 
      outPrefix = args[ 1 ];
      // get % keep argument
      try
      {
         percentKeep = Integer.parseInt( args[ 2 ] );
      }
      catch ( NumberFormatException nfe )
      {
         System.err.println( "% keep arg not a valid integer: " + args[1] );
         System.exit( 2 );
      }

      // get the size args
      int nSize = args.length - 3;
      sizeArgs = new int[ nSize ];
      
      for ( int i = 0; i < nSize; i++ )
      {
         try
         {
            sizeArgs[ i ] = Integer.parseInt( args[ i + 3 ] );
         }
         catch ( NumberFormatException nfe )
         {
            System.err.println( "size arg " + i + " is not an int: " + args[i] );
            System.exit( 3 );
         }
      }
   }
   //----------------- makeSparseData() -------------------------------------
   static void makeSparseData()
   {
      ISBoundsIterator iter = new ISBoundsIterator( inds.getBounds(), blockSize );
      Random rng = new Random( 1 );  // use a seed for predictability and testing
      ISBounds  blkBnds = new ISBounds( blockSize );
      Datum     datum = inds.createDatum();
      
      DataBlock fullBlock = inds.createDataBlock( blkBnds );
      int       fullBlockDataSize = datum.byteSize() * fullBlock.volume();
      IndexSpaceID fullBlockDims = new IntegerIndexSpaceID( blockSize );

      IndexSpaceID partialBlockDims = 
                          new IntegerIndexSpaceID( new int[ inds.dim()] );
      
      //int       fullBlockDataRefSize = 0;
      byte[]    realData = null;
      System.out.println( datum.byteSize() + " " + fullBlock.volume() );
      zeroData = new byte[ fullBlockDataSize ];
      for ( int i = 0; i < zeroData.length; i++ )
         zeroData[ i ] = 0;
      
      inBlkNum = 0;
      int outBlkNum = 0;
      while ( iter.valid() )
      {
         int ranNum = rng.nextInt( 100 );
         if ( ranNum < percentKeep )
         {
            // save the block in the sparse file
            if ( iter.isFullShape() )
            {
               inds.subblock( fullBlock, iter );
               realData = bytesFromValues( fullBlock.getValues() );
               blkSizes.add( fullBlockDims );
               System.err.println( "Write full block" + iter );
            }
            else
            {
               // create a new full size data block for each partial block, so 
               //   unfilled region is set to 0.
               DataBlock partialBlock = inds.createDataBlock( blkBnds );
               ISBounds srcBnds = ((ISBounds)iter).copy();
               ISBounds dstBnds = srcBnds.copyToOrigin();
               inds.subblock( partialBlock, dstBnds, srcBnds );
               realData = bytesFromValues( partialBlock.getValues() );
               //DataBlock partialBlock = inds.subblock( iter );
               int[] lower = partialBlock.getLowerBoundsArray();
               int[] upper = partialBlock.getUpperBoundsArray();
               for ( int d = 0; d < partialBlockDims.dim(); d++ )
                  partialBlockDims.setCoord( d, upper[d] - lower[d] + 1 );
               blkSizes.add( partialBlockDims );
               System.err.println( "Write partial block " + partialBlockDims + " " +  iter );
            }
            
            if ( realData.length != fullBlockDataSize )
               System.err.println( "ERROR: byte length discrepancy " +
                                   realData.length + " vs " + fullBlockDataSize );
            writeBlock( inBlkNum, realData );
            blkMap.put( inBlkNum, outBlkNum ); 
            blkStart.add( iter.copyLower() );
            outBlkNum++;
         }
         inBlkNum++;
         iter.next();
      }
   }
   //----------------- bytesFromValues( Values ) -------------------------
   /**
    * This functionality might best be done in Values classes. 
    * We'll do it here for now.
    * @param Values vals
    * @return ByteBuffer
    */
   static byte[] bytesFromValues( Values vals )
   {
      ByteBuffer bbuf = null;
      if ( vals instanceof FloatValues )
      {
         float[] fvals = ((FloatValues)vals ).getFloats();
         bbuf = ByteBuffer.allocate( fvals.length * 4 );
         FloatBuffer fbuf = bbuf.asFloatBuffer();
         fbuf.put(  fvals );        
      }
      else if ( vals instanceof DoubleValues )
      {
         double[] dvals = ((DoubleValues)vals ).getDataReference();
         bbuf = ByteBuffer.allocate( dvals.length * 8 );
         DoubleBuffer dbuf = bbuf.asDoubleBuffer();
         dbuf.put(  dvals );                 
      }
      else if ( vals instanceof IntValues )
      {
         int[] ivals = ((IntValues)vals ).getDataReference();
         bbuf = ByteBuffer.allocate( ivals.length * 4 );
         IntBuffer ibuf = bbuf.asIntBuffer();
         ibuf.put(  ivals );                 
      }
      else if ( vals instanceof ShortValues )
      {
         short[] svals = ((ShortValues)vals ).getDataReference();
         bbuf = ByteBuffer.allocate( svals.length * 4 );
         ShortBuffer sbuf = bbuf.asShortBuffer();
         sbuf.put(  svals );                 
      }
      else if ( vals instanceof ByteValues )
      {
         byte[] bvals = ((ByteValues)vals ).getDataReference();
         bbuf = ByteBuffer.wrap( bvals );
      }
      else if ( vals instanceof MixedValues )
      {
         byte[] bvals = ((MixedValues)vals ).getDataReference();
         bbuf = ByteBuffer.wrap( bvals );
      }
      return bbuf.array();
   }
   //----------------- writeBlock ----------------------------------------
   static void writeBlock( int outBlkNum, byte[] data )
   {
      try
      {
         outData.write( data );   // write the real data
         if ( data.length < zeroData.length )
         {
            // pad with zeros to fill the block
            outData.write( zeroData, 0, zeroData.length - data.length );
         }
      }
      catch ( IOException ioex )
      {
         System.err.println( "IO Error while writing block " + outBlkNum + 
                             ":  " + ioex.getMessage() );
         System.exit( 5 );
      }         
   }
   //----------------- makeSparseXfdl() -------------------------------------
   static void makeSparseXfdl()
   {
      // first copy the input fdlDescriptor
      FDLDescriptor copyFDL = (FDLDescriptor) inds.getFDLDescriptor().clone();
      
      // make the map data
      String[] mapInfo = new String[ blkMap.size() ];
      int k = 0;
      for ( int i = 0; i < inBlkNum; i++ )
      {
         Integer outBlk = blkMap.get( i );
         if ( outBlk != null )
         {
            String startPos = blkStart.get( outBlk ).toString();
            // toString format:  <IntISID (0,5,0)>
            //    we don't need the <IntISID or >, so remove them
            startPos = startPos.substring( 9, startPos.length() - 1 );
            String sizeData = blkSizes.get( outBlk ).toString();
            //System.err.println( startPos + "|" + sizeData );
            sizeData = sizeData.substring( 9, sizeData.length() - 1 );
            
            mapInfo[ k++ ] = ( i + // sparse block id
                                "@" + startPos + // lower position of block 
                                "=" + sizeData + // size of block: could be optional
                                "->" + blkMap.get( i )); // block id in orig ds
         }
      }
      String fileName = copyFDL.getDataFileURL();
      
      copyFDL.setDataFileURL( outPrefix + ".bin" );
      copyFDL.setSparseBlockMap( mapInfo );
      copyFDL.setBlockDimensions( new IntegerIndexSpaceID( blockSize ));
      copyFDL.writeFDL( new File ( outPrefix + ".xfdl" ));    
   }
   //------------------- openOut() -------------------------------
	protected static void openOut()
   {
		try 
      {        
         outData = new RandomAccessFile( outPrefix + ".bin", "rw" );
      } 
      catch ( IOException e )
      {
         new FileError( SDBError.Recoverable, "Couldn't Open File:" + e);
      }	
	}
	
	//------------------ closeOut() ----------------------------------
	protected static void closeOut()
   {	
		try 
      {
         outData.close();            
		} 
      catch (IOException e)
      {
			new FileError( SDBError.Recoverable, "IO Error in close()." );
		}
	}	
}
