/*
 * Comments by rdb: 02/09/12
 * 
 * Ask Bochichio about 
 *    -output file name hard coded in ARBuilder 
 *    -the code that comments out the building of data sources; will that bite
 *     us later? Can we have to different method calls?
 *    -how does this related to ARBuilder (except its call to do output)?
 * 
 * Recent history
 * 02/09/12 rdb Modified waveletMRwithError so that it does not build the 
 *              ARDataSource twice.
 * 
 */

package error;

import edu.unh.sdb.datasource.BasicBlock;
import edu.unh.sdb.datasource.DataBlock;
import edu.unh.sdb.datasource.DataCollection;
import edu.unh.sdb.datasource.DataSource;
import edu.unh.sdb.datasource.FDLDescriptor;
import edu.unh.sdb.datasource.ISBounds;
import edu.unh.sdb.datasource.ISBoundsIterator;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;
import java.util.List;

import javax.lang.model.type.ErrorType;

import ar.ARBlocksWriter;
import ar.ARBuilder;
import ar.MRHierarchy;
import ar.ARDataSource;
import ar.ARError;

import data.DataInfo;

/**
 * Builds MR Error information and then constructs and outputs the xfdl files
 * describing the tree and the ARBlocks data
 * 
 * @author svohr
 */
public class MREBuilder 
{   
    private String  fdlFilename;
    private boolean keepFDL = true;
    private boolean suppressFDL = false;
    private int     dataResolutions;   // "I've got one hundred resolutions..."
    private int     errorResolutions;
    private int     resolutionOffset;
    private int     lowestResolution;
    
    private DataBlock     source;
    private FDLDescriptor fdl;
    
    private List<ErrorFunction> errorFuncs;
    
    private String reduceMethod;
    private Coords coords;
    private Reducer [] reducers;
    private Reconstructor reconstructor;
    private LowResDataSetWriter [] dataWriters;
    private List<ErrorWriter> [] errorWriters;
            
    private DataBlock slices;
    private DataBlock reconstructed;
    private DataBlock [] low;
    
    private boolean isSlicing = false;
    
    private String filePrefix;
    //Bounds for writing, used when passing in data through memory
    private ISBounds          bounds;
    private ArrayList<String> errTypesAsString;
    private String            byteOrder;
    private BasicBlock        data;
    private float             errtol;
    private MRHierarchy       mrh;
    
    
    /**
     * Prints the usage information for the main function.
     */
    private static void printUsage ()
    {
        System.err.println(
            "USAGE:\njava selection.MREBuilder [OPTIONS] FILE\n" +
            "    --uniform\n" +
            "        Use uniform decimation to build low res (default).\n" +
            "    --wavelet\n" +
            "        Use wavelet decomposition to build low res.\n" +
            "    -n NUMBER_OF_RESOLUTIONS\n" +
            "        Default is 1\n\n" +
            "    -e NUMBER_OF_ERROR_RESOLUTIONS\n" +
            "        Default is 1\n\n" +
            "    -o ERROR_RESOLUTION_OFFSET\n" +
            "        Difference between the data resolution and its first error" +
            "        resolution. Default is 0\n\n" +
            "    --binary WIDTH HEIGHT DEPTH\n" +
            "        Specified FILE is binary, with size WIDTHxHEIGHTxDEPTH\n"+
            "        Byte order is little endian.\n\n"+
            "    --maxabs\n" +
            "        Use max absolute error (default).\n" +
            "    --avgabs\n" +
            "        Use average absolute error.\n" +
            "    --stddev\n" +
            "        Use standard deviation absolute error.\n" +
            "    --snr\n" +
            "        Use Signal-to-Noise ratio for error.\n" +
            "    --nofdl\n" +
            "        Suppress FDL files.\n" +
            "    FILE is fdl (xml) input file, unless --binary is used.\n\n" +
            "    NOTE: CLASSPATH must include '.' and 'granite.jar'\n\n");
    }
    
    
    private void parseArgs (String [] args)
    {
    	// set some defaults
        dataResolutions = 1;
        errorResolutions = dataResolutions;
        resolutionOffset = 0;
        
        errorFuncs = new ArrayList<ErrorFunction> ();
        reduceMethod = "uniform";
        
        try {
            // get the file to be read.
            if ( args.length > 0 ) {
                
                // last argument in the required FDL file.
                fdlFilename = args[args.length-1];

                // check for flags.
                for (int i=0; i<args.length; i++) { 
                    if ( args[i].equals("-n") ) {
                        dataResolutions = Integer.parseInt(args[i+1]);
                    }
                    if ( args[i].equals("-e") ) {
                        errorResolutions = Integer.parseInt(args[i+1]);
                    }                    
                    if ( args[i].equals("-o") ) {
                        resolutionOffset = Integer.parseInt(args[i+1]);
                    }

                    if ( args[i].equals("--binary") ) {
                        int width = Integer.parseInt(args[i+1]);
                        int height = Integer.parseInt(args[i+2]);
                        int depth = Integer.parseInt(args[i+3]);
                        String binfilename = args[args.length-1];

                        //System.err.printf("got size='%d %d %d'\n",
                                                    //width,height,depth);
                        writeTempFDL (binfilename, width, height, depth);
                    }
                    // error functions
                    if ( args[i].equals("--maxabs") ) {
                        errorFuncs.add( new MaxAbsErrorFunction () );
                    }
                    if ( args[i].equals("--avgabs")) {
                        errorFuncs.add( new AvgAbsErrorFunction () );
                    }
                    if ( args[i].equals("--stddev")) {
                        errorFuncs.add( new StdDevErrorFunction () );
                    }
                    if ( args[i].equals("--snr") ) {
                        errorFuncs.add( new SignalToNoiseErrorFunction () );
                    }
                    if ( args[i].equals("--uniform") ) {
                        reduceMethod = "uniform";
                    }
                    if ( args[i].equals("--wavelet") ) {
                        reduceMethod = "wavelet";
                    }
                    if ( args[i].equals("--nofdl") ) {
                        suppressFDL = true;
                    }
                    
                }
                
                // if no error functions are specified, use default.
                if ( errorFuncs.isEmpty() ) {
                    errorFuncs.add( new MaxAbsErrorFunction () );
                }

            } else {
                System.err.println ("Missing Arguments!");
                printUsage();
                System.exit(1);
            }
        } 
        catch(Exception e) {
            System.err.printf("ERROR: '%s'\n", e);
        }
        
        lowestResolution = dataResolutions 
                            + (errorResolutions - 1) 
                            + resolutionOffset;
    }
    
    private void writeTempFDL (String name, int w, int h, int d)
            throws FileNotFoundException
    {

        fdlFilename = "tmp.fdl";
        PrintWriter pw = new PrintWriter(fdlFilename);

        pw.printf(
            "<?xml version=\"1.0\" encoding=\"UTF-8\"?> " + "" +
            "<!DOCTYPE FileDescriptor " + 
                "PUBLIC \"-//SDB//DTD//EN\"  \"fdl.dtd\"> " +
            "<FileDescriptor fileName= \"%s\" " + 
                    "fileType=\"binary\" byteOrder=\"little\">" + 
                "<Bounds lower= \"0\"    upper= \"%d\" /> " + 
                "<Bounds lower= \"0\"    upper= \"%d\"/> " + 
                "<Bounds lower= \"0\"    upper= \"%d\"/> " + 
                "<Field fieldName= \"temp\" fieldType= \"float\"/> " + 
            "</FileDescriptor>", 
                  name, d-1, h-1, w-1);
        pw.flush();
        pw.close();
        keepFDL = false;
    }
    
    /* DB: No longer needed
    private void openOrigDataSource ()
    {

        fdl = FDLDescriptor.create( fdlFilename );

        try {
            source = DataSource.create( "res-zero", fdl );
        } 
        catch ( Exception ex ) {
            System.err.println( "Open error: " + ex.getMessage() );
            System.exit( -1 );
        }
        ((DataSource) source).activate();
    }
    
    */
    
    private void buildReducersAndReconstructor ()
    {
        // Build reducers, reconstructors, low res writers, and error writers.
        reducers = new Reducer [dataResolutions];
        
        
        if ( this.reduceMethod.equalsIgnoreCase("wavelet") ) {
            
            for (int i=0; i < dataResolutions; i++) 
                reducers[i] = new WaveletReducer(i+1);
            
            coords = new WaveletCoords ();
            
            reconstructor = 
                    new BasicReconstructor( new WaveletSupport (), 
                                            coords, 
                                            source.getRecordDescriptor() );
            
        }
        else {
            for (int i=0; i < dataResolutions; i++) 
                reducers[i] = new UniformDecimation(i+1);
            
            coords = new UniformCoords ();
            
            reconstructor = 
                    new BasicReconstructor( new UniformSupport (), 
                                            coords, 
                                            source.getRecordDescriptor() );
        }

        
    }
    
    private void buildWriters ()
    {
        try {
            ISBounds originalBounds = source.getBounds();
            dataWriters = new LowResDataSetWriter [dataResolutions];
            for (int i=0; i < dataResolutions; i++) {
            	if(fdl != null) {
            	dataWriters[i] = 
                    	new LowResDataSetWriter ( i+1, fdl, 
                                    	reducers[i].resultSize(originalBounds));
            	} else {
            		dataWriters[i] = 
                    	new LowResDataSetWriter ( i+1, filePrefix, 
                                    	reducers[i].resultSize(originalBounds), byteOrder);
            	}
 
                dataWriters[i].suppressFDL(suppressFDL);
            }


            errorWriters = (List<ErrorWriter> [])new List [dataResolutions];

            // This is kinda messy. For each low resolution, we need to 
            // build error files for each error resolution and for
            // each error function to be used.
            for (int d=0; d < dataResolutions; d++) {
                errorWriters[d] = new ArrayList<ErrorWriter> ();
                for (int e=0; e < errorResolutions; e++) {
                    for ( ErrorFunction f : errorFuncs ) {
                        if(fdl != null) {
                        	ErrorWriter ew = 
                                new ErrorWriter (   f, (d+1), 
                                                    (d+1+e+resolutionOffset),
                                                    lowestResolution, fdl );
                        
                        ew.suppressFDL(suppressFDL);
                        errorWriters[d].add( ew );
                        }
                        else {
                        	ErrorWriter ew = 
                        		new ErrorWriter( f, (d+1), 
                        				(d+1+e+resolutionOffset),
                        				lowestResolution, filePrefix, bounds, byteOrder );
                        	ew.suppressFDL(suppressFDL);
                            errorWriters[d].add( ew );
                            
                        }
                    }
                }
            }
        }
        catch ( Exception e ) {
             System.err.println ("Cannot open output file: " + e.getMessage() );
             System.exit(1);
            
        }
    }
        
    private void closeWriters ()
    {
        try {
            for (int i=0; i < dataResolutions; i++)
                dataWriters[i].close();


            for (int i=0; i < dataResolutions; i++) {
                for ( ErrorWriter ew : errorWriters[i] )
                    ew.close();
            }
        }
        catch ( Exception e ) {
             System.err.println ("Cannot close output file: " + e.getMessage() );
             e.printStackTrace();
             System.exit(1);
        }
    }
    
    private ISBoundsIterator buildSliceIter ()
    {
        ISBounds bounds = this.source.getBounds(); // all of the data set.
        
        // the minimum number of slices required for the lowest resolution.
        int numSlices = source.getBounds().getDimension(0);
        if (isSlicing)
        	numSlices = (int)(Math.pow(2, this.lowestResolution )) + 1;
        
        
        int [] shape = {numSlices, 
                        bounds.getDimension(1), 
                        bounds.getDimension(2)};
        int [] gaps = {-1, 0, 0};
        
        return new ISBoundsIterator (bounds, shape, gaps);
    }
    
    
    private void uniformMRwithError ()
    {
        System.err.println ("uniform");

        ISBoundsIterator sliceIter = this.buildSliceIter ();

        // build the buffers.
        
        slices = source.createDataBlock(sliceIter);
        reconstructed = source.createDataBlock(sliceIter);
        low = new DataBlock [this.dataResolutions];
        for ( int i=0; i < low.length; i++ ) { 
            low[i] = source.createDataBlock(reducers[i].resultSize(sliceIter));
        }

        // Iterate through the data set.
        for ( ; sliceIter.valid() ; sliceIter.next() ) {
            //System.err.println(sliceIter.toString());
            

            if ( ! sliceIter.isFullShape() ) {
                reconstructed = 
                    source.createDataBlock(new ISBounds (sliceIter.getDimensions()));
                slices = 
                    source.createDataBlock(new ISBounds (sliceIter.getDimensions()));
            }
            
            source.subblock(slices, sliceIter);

            // For each resolution level:
            for ( int r=1 ; r <= dataResolutions ; r += 1 ) {
            
                // If the bounds have changed, build new blocks.
                ISBounds lowSize = reducers[r-1].resultSize(sliceIter);
                if ( ! low[r-1].getBounds().equalShape(lowSize) ) {
                    low[r-1] = low[r-1].createDataBlock(lowSize);
                }

                
                // Generate low resolution.
                reducers[r-1].reduce(slices, low[r-1]);

                // Reconstruct.
                ISBounds lowBase = reducers[r-1].resultBounds(sliceIter);
                ISBounds recBase = 
                        coords.lowResToBase(0, new ISBounds (sliceIter));
                

//                System.err.println("slice: " + sliceIter.toString());
//                System.err.println("low: " + lowBase.toString());
//                System.err.println("rec: " + recBase.toString());

                reconstructor.reconstruct(  low[r-1], lowBase, r, 
                                            reconstructed, recBase );

                try {
                    // System.err.println("org: " + slices.getBounds().toString());
                    // System.err.println("rec: " + reconstructed.getBounds().toString());

                    // Write error results.
                    for ( ErrorWriter e : errorWriters[r-1] ) {
                        e.write(slices, reconstructed);
                    }
                    
                    // Write data.
                    ISBounds lowBounds = low[r-1].copyBounds();
                    ISBounds recBounds = reconstructed.copyBounds();
                    if ( sliceIter.isFullShape() ) {
                        // reduce the Z dimension by 1 to prevent overlap 
                        int [] lbound = lowBounds.copyLowerArray();
                        int [] ubound = lowBounds.copyUpperArray();
                        ubound[0] -= 1;
                        lowBounds = new ISBounds (lbound,ubound);
                        
                        lbound = recBounds.getLowerArray();
                        ubound = recBounds.getUpperArray();
                        ubound[0] -= 1;
                        
                        recBounds.setUpper(ubound);
                    }
                dataWriters[r-1].write(low[r-1], lowBounds);


                }
                catch (Exception e) {
                    System.err.println("Error writing file (" + r + "): " + e.toString());
                    System.err.println(sliceIter.toString());

                    e.printStackTrace();
                }
            }
        }

    }

    //DB: This whole method needs to be rewritten using MRHierarchy passed in by 
    //   MultiMRHierarchy
    //Error needs to be calculated and added to the array within the MRHierarchy 
    //   before sending the information to ARBlocksWriter
    private void waveletMRwithError ()
    {       
        System.err.println ("wavelet");
        
        // build the slice iterator. 
        ISBounds bounds = this.source.getBounds(); // all of the data set.
        ISBoundsIterator sliceIter = this.buildSliceIter ();
        int numSlices = sliceIter.getDimension(0);
        
        // build the buffers.
        slices = null;
        reconstructed = source.createDataBlock(sliceIter);
        low = new DataBlock [this.dataResolutions];
        for ( int i=0; i < low.length; i++ ) { 
            low[i] = source.createDataBlock(reducers[i].resultSize(sliceIter));
        }
        
        // Iterate through the data set.
        for ( ; sliceIter.valid() ; sliceIter.next() ) {

        	//System.err.println(sliceIter.toString());
            
            // Resize the iterator bounds for support values.
            ISBounds thickSlice = sliceIter.copyBounds();
            
            int [] l = thickSlice.getLowerArray();
            int [] u = thickSlice.getUpperArray();
            
            l[0] -= numSlices - 1;
            u[0] += numSlices - 2;
            
            thickSlice.setLower(l);
            thickSlice.setUpper(u);
            
            thickSlice.intersectWith(bounds);
            
            // Get the local bounds for the error window.
            ISBounds errorWindow = new ISBounds (sliceIter);
            
            l = errorWindow.getLowerArray();
            u = errorWindow.getUpperArray();
            
            l[0] -= thickSlice.getLower(0);
            u[0] -= thickSlice.getLower(0);
            
            errorWindow.setLower(l);
            errorWindow.setUpper(u);
            


            // Allocate new blocks if necessary
            if ( slices == null 
            		//|| !slices.getBounds().equalShape(thickSlice)
            		|| !sliceIter.isFullShape())  {
                reconstructed = 
                    source.createDataBlock(
                    new ISBounds (sliceIter.getDimensions()));
                slices = 
                    source.createDataBlock(
                    new ISBounds (thickSlice.getDimensions()));
            }
            

            
            source.subblock(slices, thickSlice);

            // For each resolution level:
            for ( int r=1 ; r <= dataResolutions ; r += 1 ) {
            
                // If the bounds have changed, build new blocks.
                ISBounds lowSize = reducers[r-1].resultSize(thickSlice);
                
                
                if ( ! low[r-1].getBounds().equalShape(lowSize) ) {
                    low[r-1] = low[r-1].createDataBlock(lowSize);
                }

                // Generate low resolution.
                reducers[r-1].reduce(slices, low[r-1]);

                // Reconstruct.
                ISBounds lowBase = 
                        coords.lowResToBase(0, new ISBounds (thickSlice));
                // ISBounds lowBase = reducers[r-1].resultBounds(thickSlice);
                ISBounds recBase = 
                        coords.lowResToBase(0, new ISBounds (sliceIter));

//                System.err.println("slice: " + sliceIter.toString());
//                System.err.println("low: " + lowBase.toString());
//                System.err.println("rec: " + recBase.toString());

                reconstructor.reconstruct(  low[r-1], lowBase, r, 
                                            reconstructed, recBase );

                try {
                    // System.err.println("org: " + slices.getBounds().toString());
                    // System.err.println("rec: " + reconstructed.getBounds().toString());

                    
                    // Write error results.
                    for ( ErrorWriter e : errorWriters[r-1] ) {
                        e.write(slices, errorWindow, reconstructed);
                    }
                    
                    // Write data.
                    
                    // need to resize bounds to output the same value twice.

                    ISBounds dataWindow = errorWindow.copy();
                    if ( sliceIter.isFullShape() ) {
                        // reduce the Z dimension by 1 to prevent overlap 
                        int [] lbound = dataWindow.copyLowerArray();
                        int [] ubound = dataWindow.copyUpperArray();
                        ubound[0] -= 1;
                        dataWindow = new ISBounds (lbound,ubound);
                    }

                 
                    ISBounds lowBounds = 
                            reducers[r-1].resultInData(dataWindow);

                    
                    
//System.err.println(r + " " + dataWindow.toString() + " " + lowBounds.toString());
                    
                    lowBounds = lowBounds.intersection (low[r-1].getBounds());
                    if ( lowBounds != null ) {
                        //DB - construct the MRHierarchy errors
                        mrh.setError(reconstructed, r-1);
                        
                     
                       dataWriters[r-1].write(low[r-1], lowBounds);
                    }

                }
            	
            
                catch (Exception e) {
                    System.err.println("Error writing file (" + r + "): " + e.toString());

                    System.err.println(sliceIter.toString());

                    e.printStackTrace();
                }
            }
            
            // build ARDataSource and output xml structure file
            
            // rdb: it looks like we build the ARDataSource twice, once in
            //    bldr.processData and once in arbw.processData
            //    
            //ARBuilder bldr = new ARBuilder();
            //bldr.processData( mrh, errtol ); 
            //ARBlocksWriter arbw = new ARBlocksWriter();
            //arbw.processData( mrh, errtol );
            
            // Only need the following (once I added getDataSource() to ARBuilder
            //ARDataSource ards = bldr.getDataSource();

            //But then I realized the bldr.processData doesn't do anything
            //  except create the ards and call arwriter to print it.
            
            ARError tolerance = new ar.ARFloatError( errtol );            
            ARDataSource ards = new ARDataSource( mrh );
            ards.writeARfdl( "testsend0-ar" ); // where should this come from
            ards.dumpBlocks( tolerance );  
        }
    }
        
    private void setup() {
    	// set some defaults
        resolutionOffset = 0;
        
        errorFuncs = new ArrayList<ErrorFunction> ();
        reduceMethod = "wavelet";
        
        for(String s: errTypesAsString) {
        	
        	if( s.equalsIgnoreCase("maxabs") )
        		errorFuncs.add( new MaxAbsErrorFunction () );
        	else if( s.equalsIgnoreCase("avgabs") )
        		errorFuncs.add( new MaxAbsErrorFunction () );
        	else if( s.equalsIgnoreCase("stddev") )
        		errorFuncs.add( new MaxAbsErrorFunction () );
        	else if( s.equalsIgnoreCase("snr") )
        		errorFuncs.add( new MaxAbsErrorFunction () );
        
        }
    }
    
     public void go (MRHierarchy mrh, ISBounds[] bounds, String fdlPrefix, 
    		DataInfo di ) 
    {
       if ( mrh == null )
       {
//        	this.openOrigDataSource();
          isSlicing = true;
       }
       else {
        	filePrefix = fdlPrefix;
        	isSlicing = false;
        	errTypesAsString = di.getErrorTypes();
        	this.byteOrder = di.getByteOrder();
    		dataResolutions = di.getNumRes();
    		errorResolutions = di.getErrRes();
    		this.errtol = di.getErrTol();
    		this.mrh = mrh;
    		setup();
        }
        for(int i = 0; i < mrh.getData().length; i++) {
        	source = mrh.getData()[i];
        	this.bounds = bounds[i];
        	this.buildReducersAndReconstructor();
            
            this.buildWriters();
//

            if ( reduceMethod.equalsIgnoreCase("wavelet") )
                this.waveletMRwithError();
            else
                this.uniformMRwithError();

            // close all files.
            this.closeWriters();
        }
        System.out.println("Finished");
    }
     /**
      * @deprecated 
      * @param mrh
      * @param bounds
      * @param fdlPrefix
      * @param errTypes
      * @param byteOrder
      * @param numRes
      * @param errRes
      * @param errTol 
      */
    public void go (MRHierarchy mrh, ISBounds[] bounds, String fdlPrefix, 
    		ArrayList<String> errTypes, String byteOrder, int numRes, 
    		int errRes, float errTol) 
    {
        // Setup everything before running through the data source.
        
        // Open original data set.
        //Take this out once new code is working and make
    	//else statement the only batch of code to execute
    	if ( mrh == null) {
//        	this.openOrigDataSource();
        	isSlicing = true;
        }
        else {
        	filePrefix = fdlPrefix;
        	isSlicing = false;
        	errTypesAsString = errTypes;
        	this.byteOrder = byteOrder;
    		dataResolutions = numRes;
    		errorResolutions = errRes;
    		this.errtol = errTol;
    		this.mrh = mrh;
    		setup();
        }
        for(int i = 0; i < mrh.getData().length; i++) {
        	source = mrh.getData()[i];
        	this.bounds = bounds[i];
        	this.buildReducersAndReconstructor();
            
            this.buildWriters();
//

            if ( reduceMethod.equalsIgnoreCase("wavelet") )
                this.waveletMRwithError();
            else
                this.uniformMRwithError();

            // close all files.
            this.closeWriters();
        }
        System.out.println("Finished");
    }
    
    public static void main ( String [] args )
    {
        MREBuilder builder = new MREBuilder ();
        
        builder.parseArgs (args);
        
        //Make 2 go methods eventually

        
        if ( !builder.keepFDL )
            new File(builder.fdlFilename).delete();
    }
    


}
