/*
 * Wavelet.java
 *
 * Created on May 30, 2009, 10:12 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package error;

/**
 * This implementation of n-dim wavelet composition, generates the final
 * result for each data point at one time, rather than doing d separate 1-dim
 * wavelet operations. That means we need an n^d array of coefficients where
 * n is the support of the wavelet. E.g., the Haar 3D wavelet has 2x2x2 factor
 * array where each entry is 1/8. D4 wavelet has a 4x4x4 factor array where 
 *    f(s,r,c) = coeff(s) * coeff(r) * coeff( c ) / divisor
 * where coeff( i ) is the i'th wavelet coefficient of the D4 wavelet for 
 *    i = 0, 1, 2, 3
 * and 
 *    divisor = sqrt( 2 ) ^ dim
 *
 * @author rdb
 */
import edu.unh.sdb.datasource.*;
import edu.unh.sdb.common.*;

public class Wavelet
{
   //------------------ class variables --------------------------
   //------------------ instance  variables ----------------------
   // The coefficient arrays are the basic wavelet coefficients, one per
   //  support
   protected float[] sumCoeff; // wavelet coeff for computing summary weights
   protected float[] detCoeff; // wavelet coeff for computing detail weights
   
   // The factor arrays represent the values multiplied by each data value
   //   in the n-dim array for the wavelet factors. These arrays have
   //   dim * support entries
   protected float[] sumFactors;
   protected float[] detFactors;
   
   
   protected float   divisor;     // = sqrt( 2 ) ** n where n is dimension
   protected int     support;     // same as sumFactors.length
   protected int     dim;         // dimensionality of data
   
   //------------------ constructors -----------------------------
   /** Creates a new instance of Wavelet */
   public Wavelet( int d )
   {
      dim     = d;
      divisor = (float) Math.pow( Math.sqrt( 2 ), d );
   }
   //-------------- computeFactors() --------------------------------------
   /**
    * once children classes have filled in the sumCoeff and detCoeff arrays,
    * can compute the factors in common.
    */
   protected void computeFactors()
   {
      sumFactors = new float[ (int) Math.pow( support, dim ) ];
      detFactors = new float[ (int) Math.pow( support, dim ) ];
      
      switch ( dim )
      {
         case 1: computeFactors1D(); break;
         case 2: computeFactors2D(); break;
         case 3: computeFactors3D(); break;
         default: 
            //System.err.println( "Can't do wavelet in " + dim + " dimensions." );
            break;
      }
   }
   //----------------- computeFactors1D() ------------------------------
   /** 
    * compute the factors arrays for 1D
    */
   private void computeFactors1D()
   {
      for ( int i = 0; i < support; i++ )
      {
         sumFactors[ i ] = sumCoeff[ i ];
         detFactors[ i ] = detCoeff[ i ];
      }
   }
   //----------------- computeFactors2D() ------------------------------
   /** 
    * compute the factors arrays for 2D
    */
   private void computeFactors2D()
   {
      //int i = 0;
      for ( int r = 0; r < support; r++ )
      {
         for ( int c = 0; c < support; c++ )
         {
            sumFactors[ r+c ] = sumCoeff[ r ] * sumCoeff[ c ];
            detFactors[ r+c ] = detCoeff[ r ] * detCoeff[ c ];
         }
      }
   }
   //----------------- computeFactors2D() ------------------------------
   /** 
    * compute the factors arrays for 2D
    */
   private void computeFactors3D()
   {
      int i = 0;
      
      assert (divisor > 0);
      for ( int s = 0; s < support; s++ )
      {
         for ( int r = 0; r < support; r++ )
         {
            for ( int c = 0; c < support; c++ )
            {
               sumFactors[ i ] = sumCoeff[ s ] * sumCoeff[ r ] * sumCoeff[ c ] 
                                   / divisor;
               //System.out.print( i + " " + sumFactors[ i ] + " " );
               detFactors[ i ] = detCoeff[ s ] * detCoeff[ r ] * detCoeff[ c ]
                                   / divisor;
               
               assert (sumFactors[i] == sumFactors[i]);
               
               i += 1;
            }
            //System.out.println();
         }
         //System.out.println( "--" );
      }
   }
   
   //------------ doSummary( DataCollection, int, DataBlock ) ---------------
   /**
    * apply one step of wavelet to the input DataCollection and return
    * a DataBlock.
    */
   public DataBlock doSummary( DataCollection inData, int attr )
   {
      // Make a DataBlock to return the summary data
      
      ISBounds inBounds = inData.getBounds();
      // need output bounds to be 1/2 size of the input in all dimensions
      // first get the sizes of the input
      int[]  outSizes = inBounds.getDimensionsArray();
      
      // now halve them -- for odd-length input, we'll just lose the end,
      //        but thats ok
      for ( int i = 0; i < inData.dim(); i++ )
         outSizes[i] = outSizes[i] / 2;
      // now make a bounds for the output
      ISBounds outBounds = new ISBounds( outSizes );
      
      // get a RecordSpec to extract just the right attribute
      RecordSpec rs = new RecordSpec( new int[] { attr } );
      
      // get an output block with the right bounds and just one attribute
      DataBlock outBlock  = inData.createDataBlock( outBounds, rs );
      
      doSummary( inData, inBounds, attr, outBlock );
      return outBlock;
   }
   //-------- doSummary( DataCollection, ISBounds, int, DataBlock ) ------------
   /**
    * apply one step of wavelet to the portion of the input DataCollection 
    * specified by the bounds and return the result in the DataBlock that is 
    * passed in. If the bnds is smaller than the DataBlock, only return that
    * portion of the bounds in the DataBlock; set the remainder of the 
    * DataBlock to 0.
    */
   public void doSummary( DataCollection inData, ISBounds bnds, 
                          int attr, DataBlock outBlock )
   {
      // need output bounds to be 1/2 size of the input in all dimensions
      // first get the sizes of the input
      ISBounds outBounds = outBlock.getBounds();
      
      // get a RecordSpec to extract just the right attribute
      RecordSpec rs = new RecordSpec( new int[] { attr });
            
      // waveBasis must be large enough in each dimension to hold the points
      //  needed for the particular kind of wavelet. 2 for Haar, 4 for D4, etc.
      // Since we need to step by 2 all the time, the gaps array must be 
      //  negative number that is 2 - support.
      int[]    waveBasis = new int[ inData.dim() ];
      int[]    gaps      = new int[ inData.dim() ];
      for ( int i = 0; i < waveBasis.length; i++ )
      {
         waveBasis[ i ] = support;
         gaps[ i ] = 2 - support;
      }
      
      ISBounds waveBounds = new ISBounds( waveBasis );
      
      // And pre-allocated a DataBlock to take the nxnx ... x n subblock
      DataBlock waveBlock = inData.createDataBlock( waveBounds, rs );
      
      // We need a block iterator that will go through the data space, but it
      //   will always step by 2 in each dimension. Hence, we need a negative
      //   gaps parameter
      ISBoundsIterator inIter = new ISBoundsIterator( bnds, waveBasis, gaps );
      
      // We also need an iterator over the output space, but this is a simple
      // iterator that just identifies one point at a time (not a block).
      ISIterator outIter = new ISIterator( outBounds );
      
      // pre-allocate an array to fill with the data
      float[] vals = new float[ waveBlock.volume() ];
      
      //System.out.println( "blocksize: " + vals.length );
      
      for ( ; inIter.valid(); inIter.next(), outIter.next() )
      {
         inData.subblock( waveBlock, inIter, rs );  // get the next block
         waveBlock.getFloats( vals );
         
         // now compute the wavelet transform for the block
         float value = 0.0f;
         for ( int i = 0; i < vals.length; i++ )
            value += sumFactors[ i ] * vals[ i ];
         outBlock.setFloat( outIter, 0, value ); // store result
      }
   }
}
