/*
 * This class combines the functionality of building data set files (.dtd) 
 * and file definition file (.fdl). 
 */

package selection;

import edu.unh.sdb.common.Datum;
import edu.unh.sdb.datasource.DataCollection;
import edu.unh.sdb.datasource.FDLDescriptor;
import edu.unh.sdb.datasource.ISBounds;
import edu.unh.sdb.datasource.ISIterator;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;

/**
 * Allows the writing of both a file containing a data set and the corresponding
 * FDL meta data file.
 * 
 * @author Sam Vohr
 */
public class DataSetWriter 
{
    
    private String filename;
    private FDLDescriptor fd;
    
    private FileOutputStream dataOut;
    private XFDLWriter fdlOut;
    
    /**
     * Flag for suppressing the generation of the FDL meta data file.
     */
    private boolean FDLEnabled;
    
    /**
     * creates a new instance and opens the .dtd file for writing. 
     * @param name name of the data set to create .dtd and .fdl files.
     * @param fdlDesc the FDLDescriptor for the new data set.
     */
    public DataSetWriter (String name, FDLDescriptor fdlDesc) 
            throws FileNotFoundException 
    {
        this.filename = name;
        this.fd = fdlDesc;
        
        this.fd.setDataFileURL( this.filename + ".dtd" );
        
        this.dataOut = new FileOutputStream ( this.filename + ".dtd" );
        this.fdlOut = new XFDLWriter ();
    }
    
    /**
     * Writes the contents of the Datum to the data file. Assumes the Datum's
     * contents is a single float.
     * 
     * @param d Datum to write out
     * 
     * @throws java.io.IOException
     */
    public void write (Datum d) throws IOException
    {
        int atts = d.getNumAttributes();
        
        
        float value;
        
        for (int i=0; i < atts; i++) {
            value = d.getFloat(0);
            this.write ( value );
        }

    }
    
    /**
     * Writes a single float to the data file.
     * 
     * @param f a floatingpoint value
     * 
     * @throws java.io.IOException
     */
    public void write (float f) throws IOException
    {
        this.write( floatToBytes(f) );
    }
    
    /**
     * Writes an array of bytes to the data file.
     * 
     * @param inBytes bytes to write
     * 
     * @throws java.io.IOException
     */
    public void write (byte [] inBytes) throws IOException
    {
        byte outBytes [] = inBytes;
        
/*      System.err.print ("in bytes ");
        for (int i=0; i<inBytes.length; i++) {
            System.err.print(Byte.toString(inBytes[i]) + " ");
        }
        System.err.println(); */
        
        // depending on the endianess of the system, output the data.
        if ( fd.isBigEndian() ) {
            this.dataOut.write( outBytes );
        }
        else {
            outBytes = new byte[ inBytes.length ];
            for (int i=0, j = inBytes.length - 1; i < inBytes.length; i++, j--) {
                outBytes[j] = inBytes[i];
            }
            this.dataOut.write( outBytes );
        }
        
/*      System.err.print ("out bytes ");
        for (int i=0; i<outBytes.length; i++) {
            System.err.print(Byte.toString(outBytes[i]) + " ");
        }
        System.err.println(); */
        
    }
    
    /**
     * 
     */
    public void write ( DataCollection inData, ISBounds bounds )
    {
        ISIterator iter = new ISIterator( bounds );
        
        Datum dat = inData.createDatum();

        // go through the bounds, and output 
        for ( ; iter.valid(); iter.next() ) {
            try {
                inData.datum(dat, iter);
                this.write(dat);
                
            } catch (Exception e) {
                System.err.println ("Error in writing file " + this.filename);
                System.exit(1);
            }
        }
    }
    
    /**
     * Closes the data file and outputs the meta data file.
     * 
     * @throws java.io.IOException
     */
    public void close () throws IOException 
    {
        // close the data stream.
        this.dataOut.close();
        
        // write the fdl file
        if ( FDLEnabled ) {
            this.fdlOut.setOutputFile (this.filename + ".fdl");
            this.fdlOut.writeFDL (fd);
            this.fdlOut.close();
        }
        
    }
    
    /**
     * Sets the flag to output meta data when the DataSetWriter is closed.
     * 
     * @param b boolean value
     */
    public void suppressFDL (boolean b)
    {
        this.FDLEnabled = !b;
    }
    
    /**
     * utility function to converting a float into a byte array. Byte order is 
     * preserved.
     * 
     * @param f a floating point value
     * @return a byte array for the specified float.
     */
    private byte [] floatToBytes (float f) 
    {
        int i = Float.floatToIntBits(f);
        
        byte [] array = new byte [4];
        
        array[0] = (byte)((i&0xff000000) >> 24);
        array[1] = (byte)((i&0x00ff0000) >> 16);
        array[2] = (byte)((i&0x0000ff00) >> 8);
        array[3] = (byte)((i&0x000000ff) >> 0);
        
        return array;
    }
    
}
