/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package selection;

import edu.unh.sdb.common.RecordSpec;
import edu.unh.sdb.datasource.DataBlock;
import edu.unh.sdb.datasource.DataCollection;
import edu.unh.sdb.datasource.ISBounds;
import edu.unh.sdb.datasource.ISIterator;

/**
 *
 * @author svohr
 */
public class ErrorRegionReconstructor implements Reconstructor {
    
    private int dataResolution = 1;
    private int pointDist;
    private ErrorRegionInterpolator interpolator = null;
    
    public ErrorRegionReconstructor ( int dataResLevel )
    {
        dataResolution = dataResLevel;
        pointDist = (int) Math.pow(2, dataResolution);
        interpolator = new ErrorRegionInterpolator ();
    }
    
    public float [][][] reconstruct (float [][][] original)
    {
        int dimz = original.length;
        int dimy = original[0].length;
        int dimx = original[0][0].length;
        
        float [][][] result = new float [dimz][dimy][dimx];
        
        for ( int i=0; (i + pointDist) < dimz; i += pointDist ) {
            for ( int j=0; (j + pointDist) < dimy; j += pointDist ) {
                for ( int k=0; (k + pointDist) < dimx; k += pointDist ) {
                    interpolator.interpolate(original, result, i, j, k, 
                            i+pointDist, j+pointDist, k+pointDist);
                }
            }
        }
        
        return result;
    }

    /**
     * Reconstructs a volume using a low resolution vesion and an interpolation 
     * method.
     * 
     * @param inData        the low-resolution data
     * @param base          the bounds of the low resolution in base coordinates
     * @param resolution    the resolution level
     * @param region        the bounds of the region to reconstruct
     * @param attr          the attribute to extract
     * @return              the reconstructed data block
     */
    public DataBlock reconstruct( DataCollection inData, ISBounds base, 
                                  int resolution, ISBounds region, int attr ) 
    {
        // Make a DataBlock to return the reconstructed volume

        // need output bounds to be 1/2 size of the input in all dimensions
        // first get the sizes of the input
        int [] outSizes = region.getDimensionsArray();

        // now make a bounds for the output
        ISBounds outBounds = new ISBounds( outSizes );
              
        // get a RecordSpec to extract just the right attribute
        RecordSpec rs = new RecordSpec( new int[] { attr });
        
        // get an output block with the right bounds and just one attribute
        DataBlock outBlock  = inData.createDataBlock( outBounds, rs );
        
        
        // for each point in the new block...
        ISIterator outIter = new ISIterator( outBounds );
        int [] index = new int [3];
        float reconstructedValue = 0;
        
        for ( int i=0; i < outSizes[2]; i++ ) {
            for ( int j=0; j < outSizes[1]; j++ ) {
                for ( int k=0; k < outSizes[0]; k++, outIter.next() ) {
                    
                    // local coords of the point to reconstuct
                    index[0] = i; 
                    index[1] = j; 
                    index[2] = k;
                    
                    // for each point in the out set, find the points that
                    // support it.
                    ISBounds support = supportForPointAtRes(index, resolution);
                    DataBlock d = inData.subblock(support);
                    
                    // move bounds to world base coords
                    support = lowResToBaseCoords(resolution, support);
                    support.move(base.getLower());
                    
                    // move point to world base coords
                    for (int l=0; l<index.length; l++) {
                        index[l] += region.getLower(i);
                    }
                    index = lowResToBaseCoords(0, index);
                    
                    // interpolate the value
                    reconstructedValue =
                            this.interpolator.interpolate(d, base, index);
                    
                    outBlock.setFloat( outIter, 0, reconstructedValue );
                }
            }
        }
        
        return outBlock;
    }

    private ISBounds supportForPointAtRes ( int [] point, int res ) 
    {
        int [] lower = new int[3];
        int [] upper = new int[3];
        
        for (int i=0; i<point.length; i++) {
            lower[i] = point[i] / (int)(Math.pow(2.,res));
            upper[i] = lower[i] + 1;
        }
        
        return new ISBounds (lower, upper);
    }
    
    private int [] lowResToBaseCoords ( int res, int [] point )
    {
        for (int i=0; i<point.length; i++) {
            point[i] *= (int)(Math.pow(2, res+1));
        }
        return point;
    }

    private ISBounds lowResToBaseCoords ( int res, ISBounds b )
    {
        int [] lower = b.getLowerArray();
        int [] upper = b.getLowerArray();
        
        lower = lowResToBaseCoords (res, lower);
        upper = lowResToBaseCoords (res, upper);
        
        b.setLower(lower);
        b.setUpper(upper);
        
        return b;
    }
    
}
