/*
 * 
 */

package selection;

import edu.unh.sdb.common.Datum;
import edu.unh.sdb.common.RecordDescriptor;
import edu.unh.sdb.datasource.AxisOrdering;
import edu.unh.sdb.datasource.DataBlock;
import edu.unh.sdb.datasource.FDLDescriptor;
import edu.unh.sdb.datasource.ISBounds;
import edu.unh.sdb.datasource.ISBoundsIterator;
import edu.unh.sdb.datasource.ISIterator;
import java.io.FileNotFoundException;
import java.io.IOException;

/**
 * Calculates and writes the error values for a set of slices given a resolution
 * level.
 * 
 * @author svohr
 */
public abstract class ErrorWriter  {

    private int resLevel;
    private int errorResLevel;
    private int maxSlices;
    private int regionSize;
    
    private DataSetWriter errorFile;
    private Reconstructor reconstuctor; 
    
    /**
     * Creates an ErrorWriter object.
     * 
     * @param name name of the data set. Error files will be "name_error" 
     *        .dtd and .fdl
     * @param level resolution level. 0 is base resolution.
     * @param size ISBounds of the dimensions of the error set.
     * @param base FDLDescriptor of the base resolution (for Byte Order)
     * 
     * @throws java.io.FileNotFoundException
     */
    public ErrorWriter (String name, int level, int errorLevel, int maxlevel,
            ISBounds size, FDLDescriptor base) 
            throws FileNotFoundException
    {
        this.resLevel = level;
        this.errorResLevel = errorLevel;
        this.maxSlices = (int) (Math.pow(2, maxlevel) + 1);
        this.regionSize = (int) (Math.pow(2, errorLevel) + 1);
        
        FDLDescriptor fd = new FDLDescriptor (  new RecordDescriptor ("f"), 
                                                size.copy(),
                                                new AxisOrdering (3) );

        fd.setByteOrder( base.getByteOrder() );
        this.errorFile = new DataSetWriter (name + "_" + errorLevel, fd);
        this.reconstuctor = new ErrorRegionReconstructor (level);
    }
    
    /**
     * Sets the flag to output meta data when the DataSetWriter is closed.
     * 
     * @param b boolean value
     */
    public void suppressFDL (boolean b)
    {
        errorFile.suppressFDL(b);
    }
    
    /**
     * Calculates the error values for the regions located with the Slice array.
     * This is acheived by...
     *  1) isolating the error regions 
     *  2) reconstructing each region based on the points that will be available
     *  3) comparing the reconstructed copy with the base resolution
     * 
     * @param slices DataBlock array containing the original data.
     * 
     * @throws java.io.IOException
     */
    public void write (DataBlock slices) throws IOException 
    {
        int count = 0;
        int vcount = 0;
        
      //System.err.println ("Error Write called... res " + resLevel);
        
        // a cell slice...
        int [] shape = {regionSize, regionSize, regionSize};
        int [] gaps = {-1, -1, -1};
        float [][][] cell = new float [regionSize][regionSize][regionSize];
        
        int [] dims = slices.getDimensionsArray();
        int zLength = dims[0];
        int yLength = dims[1];
        int xLength = dims[2];
                    
        // create the cell iterator
        ISBoundsIterator cellIter = 
                new ISBoundsIterator( slices.getBounds(), shape, gaps );
        
        int maxRegions = getMaxRegion(xLength,yLength,zLength);

        
        // repeat for all error cells contained in the slices.
        for (int i=0; i < maxRegions && cellIter.valid(); i++, cellIter.next())
        {
            
            // copy the cell data from the cell
            ISIterator points = 
                    new ISIterator ( cellIter.copyBounds() );

            Datum d = slices.createDatum();

            int dimz = cellIter.getDimension(0);
            int dimy = cellIter.getDimension(1);
            int dimx = cellIter.getDimension(2);

            // copy the data
            for ( int j=0 ; j < dimz; j++ ) {
                for (int k=0 ; k < dimy; k++) {
                    for (int l=0 ; l < dimx; l++) {
                        slices.datum(d, points);
                        cell[j][k][l] = d.getFloat(0);
                        points.next();
                    }
                    // fill in the rest of the row.
                    for ( int l=dimx; l < regionSize ; l++ ) {
                        cell[j][k][l] = cell[j][k][dimx-1];
                    }
                }
                // fill in the missing rows.
                for ( int k=dimy ; k < regionSize ; k++ ) {
                    for ( int l=0; l < regionSize; l++ )
                        cell[j][k][l] = cell[j][dimy-1][l];
                }
            }
            // fill in the missing slices.
            for (int j=dimz; j<regionSize; j++) {
                for (int k=0 ; k < regionSize; k++) {
                    for ( int l=0; l < regionSize; l++ ) {
                        cell[j][k][l] = cell[dimz-1][k][l];
                    }
                }
            }

            // replaces the values at the corner so interpolation can 
            // be done.   
            if ( !cellIter.isFullShape() ) 
                cell = fixCorners (cell, dimx, dimy, dimz);

            // reconstruct the cell by interpolating the 
            // points at the corners
            float [][][] cellReconstructed = 
                    reconstuctor.reconstruct (cell);

            // get the value for this cell
            float errorValue = calculate (cell, cellReconstructed);

            // write the result
            // System.err.println (errorValue);
            errorFile.write( errorValue );

               

        }
        // done.
    }
    
    /**
     * Closes the ErrorWriter
     * 
     * @throws java.io.IOException
     */
    public void close () throws IOException {
        errorFile.close();
    }

    public abstract float calculate 
            (float [][][] actual, float [][][] reconstructed);
    
    private float [][][] fixCorners (float [][][] cell, 
            int dimx, int dimy, int dimz)
    {
        /*      E
         *    / | \
         *  A   G   F 
         *  | X   X |
         *  C   B   H
         *    \ | /
         *      D
         * 
         * A is given (error cell cannot exist with out it)
         * B, E, C are copied from A if they do not exist.
         * D, F, G are copied from B|C|A, E|B|A, C|E|A respectively.
         * H   
         * 
         */
        
        int end = regionSize - 1;
        
        // B
        if (dimx < regionSize) 
            cell[0][0][end] = cell[0][0][0];
        
        // C
        if (dimy < regionSize)
            cell[0][end][0] = cell[0][0][0];
        
        // E
        if (dimz < regionSize)
            cell[end][0][0] = cell[0][0][0];
        
        // D
        if ( dimx == regionSize && dimy < regionSize )
            cell[0][end][end] = cell[0][0][end];
        else if ( dimx < regionSize && dimy == regionSize )
            cell[0][end][end] = cell[0][end][0];
        else if ( dimx < regionSize && dimy < regionSize )
            cell[0][end][end] = cell[0][0][0];
        
        // F
        if ( dimx == regionSize && dimz < regionSize )
            cell[end][0][end] = cell[0][0][end];
        else if ( dimx < regionSize && dimz == regionSize )
            cell[end][0][end] = cell[end][0][0];
        else if ( dimx < regionSize && dimz < regionSize )
            cell[end][0][end] = cell[0][0][0];
        
        // G
        if ( dimy == regionSize && dimz < regionSize )
            cell[end][end][0] = cell[0][end][0];
        else if ( dimy < regionSize && dimz == regionSize )
            cell[end][end][0] = cell[end][0][0];
        else if ( dimy < regionSize && dimz < regionSize )
            cell[end][end][0] = cell[0][0][0];
        
        // H
        if ( dimx == regionSize && dimy == regionSize)
            cell[end][end][end] = cell[0][end][end];
        else if ( dimx == regionSize && dimz == regionSize )
            cell[end][end][end] = cell[end][0][end];
        else if ( dimy == regionSize && dimz == regionSize)
            cell[end][end][end] = cell[end][end][0];
        else if ( dimy == regionSize && dimz < regionSize && dimx < regionSize)
            cell [end][end][end] = cell[0][end][0];
        else if ( dimx == regionSize && dimz < regionSize && dimy < regionSize)
            cell [end][end][end] = cell[0][0][end];
        else if ( dimz == regionSize && dimx < regionSize && dimy < regionSize)
            cell [end][end][end] = cell[end][0][0];
        else
            cell [end][end][end] = cell[0][0][0];
            

        return cell;
    }

    private int getMaxRegion(int xLength, int yLength, int zLength) {
                
        int dist = (int)Math.pow(2, errorResLevel);
        
        int xdim = xLength / dist;
        int ydim = yLength / dist;
        int zdim = maxSlices / dist;
        
        if ( xLength % dist != 0 ) 
            xdim += 1;
        if ( yLength % dist != 0 ) 
            ydim += 1;

        
        return xdim * ydim * zdim;
    }
    
    
}
