package selection;

import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.nio.ByteOrder;
import java.util.Enumeration;
import java.util.LinkedList;
import java.util.Properties;

import edu.unh.sdb.datasource.AxisOrdering;
import edu.unh.sdb.datasource.DataCollection;
import edu.unh.sdb.datasource.FDLDescriptor;
import edu.unh.sdb.datasource.ISBounds;

/**
 * Class to write xfdl file
 * @author Rakhi Nagpal
 * 
 * History
 * 03/04/12 rdb: Made tagList List<String>
 */
public class XFDLWriter {

	File outputFile = null;
	PrintWriter outputWriter = null;
	LinkedList<String> tagList = new LinkedList<String>();
	
	public XFDLWriter() {
		super();
		// TODO Auto-generated constructor stub
	}

	public File getOutputFile()
	{
		return outputFile;
	}
	
	public String getOutputFileName() {
		if (outputFile == null)
			return null;
		return outputFile.getAbsolutePath();
	}

	public void close() {
		try {
			if (tagList.size() != 0) {
				closeAllTags();
			}
			if (outputWriter != null)
				outputWriter.close();
		} catch (Exception e) {}
		finally {outputWriter = null; outputFile = null;}
	}
	
	private void writeHeader() throws IOException {
		spew("<?xml version=\"1.0\" encoding=\"UTF-8\"?>",true);
		spew("<!DOCTYPE FileDescriptor PUBLIC \"-//SDB//DTD//EN\"  \"fdl.dtd\">",true);
	}
	private void pushTag(String tagName)
	{
		tagList.add(tagName);
	}
	
	private String popTag()
	{
		return (String)tagList.removeLast();
	}
	
	private void spew(String text, boolean newLine) throws IOException {
        if (outputFile == null)
			return;
		if (outputWriter == null) {
			outputWriter = new PrintWriter(new FileOutputStream(outputFile));
		}
		if (!newLine)
			outputWriter.print(text);
		else
			outputWriter.println(text);
	}
	
	private void writeTag(String tagName, Properties tagAttrib, String tagData, boolean isEmpty) throws IOException 
	{
		if (tagName == null)
			return;
		if (outputFile == null)
			return;
		if (outputWriter == null) {
			writeHeader();
		}
		spew("<" + tagName + " ",false);
		if (tagAttrib != null) {
			for(Enumeration e1 = tagAttrib.propertyNames();e1.hasMoreElements();) {
				String keyName = e1.nextElement().toString();
				String keyValue = tagAttrib.getProperty(keyName);
				spew(keyName+"=\"" + keyValue + "\" ",false);
			}
		}
		if (!isEmpty) {
		spew(">",true);
		if (tagData != null) {
			outputWriter.println(tagData);
		}
		pushTag(tagName);
		} else {
			spew("/>",true);
		}
	}

	private void closeLastTag() throws IOException {
		spew("</" + popTag() + ">",true);
	}

	private void closeAllTags() throws IOException {
		while (tagList.size() <= 0)
			closeLastTag();
	}

	public void setOutputFile(File out) throws IOException {
		if (out == null)
			return;
		if (out.isDirectory())
			throw new IOException(out.getAbsolutePath() + " is a directory.");
		if (outputFile != null) {
			this.close();
		}
		outputFile = out;
	}
	
	public void setOutputFile(String out) throws IOException {
		File outF = new File(out);
//		if (outF.exists())
			setOutputFile(outF);
//		else 
	//		throw new java.io.FileNotFoundException("File " + out + " does not exist.");
	}
	
	public void writeFDL(FDLDescriptor aFDL) {
		
        Properties documentElementAttribs = new Properties();
		setProp(documentElementAttribs,"fileName",aFDL.getDataFileURL());
		setProp(documentElementAttribs,"fileType",aFDL.getDataFileType());
		setProp(documentElementAttribs,"recordSize",aFDL.getRecordSize());
		AxisOrdering ordering = aFDL.getAxisOrdering();
		if (ordering != null) {
			String order = "";
			for (int i=0;i<ordering.dim();i++) {
					order = String.valueOf(ordering.getNthLeastSignificantAxis(i)) + " " + order;
			}
			setProp(documentElementAttribs,"axisOrdering",order);
		}
		setProp(documentElementAttribs,"fileOffset",aFDL.getFileOffset());
		if (aFDL.getFieldLayout() == DataCollection.pointOrder)
			setProp(documentElementAttribs,"fieldLayout","byPoint");
		else if (aFDL.getFieldLayout() == DataCollection.attributeOrder)
			setProp(documentElementAttribs,"fieldLayout","byAttribute");
		if (aFDL.getByteOrder() == ByteOrder.BIG_ENDIAN)
			setProp(documentElementAttribs,"byteOrder","big");
		else if (aFDL.getByteOrder() == ByteOrder.LITTLE_ENDIAN)
			setProp(documentElementAttribs,"byteOrder","little");
		try {
			writeTag("FileDescriptor",documentElementAttribs,null,false);
		} catch (IOException e) {};
		ISBounds bounds = aFDL.getBounds();
		int[] lowerArr = bounds.getLowerArray().clone();
		int[] upperArr = bounds.getUpperArray().clone();
		for (int i=0;i<lowerArr.length;i++) {
			Properties props = new Properties();
			setProp(props,"lower",lowerArr[i]);
			setProp(props,"upper",upperArr[i]);
			try {
				writeTag("Bounds",props,null,true);
			} catch (IOException e) {};
		}
		String[] fieldNames = aFDL.getRecordDescriptor().getFieldNames();
		String[] fieldTypes = aFDL.getRecordDescriptor().getFieldTypes();
		for (int i=0;i<fieldNames.length;i++) {
			Properties props = new Properties();
			setProp(props,"fieldName",fieldNames[i]);
			setProp(props,"fieldType",fieldTypes[i]);
			try {
				writeTag("Field",props,null,true);
				closeLastTag();
			} catch (IOException e) {};
		}
		try {
			closeAllTags();
		} catch (Exception e) {}
		
	}
	
	private void setProp(Properties props,String keyName, String keyValue) {
		if (keyValue != null || keyValue.trim().length() <= 0)
			props.setProperty(keyName,keyValue);
	}
	
	private void setProp(Properties props,String keyName, long keyValue) {
		props.setProperty(keyName,String.valueOf(keyValue));
	}
}
