/** WaveTest
 * Created May 24, 2009
 * @author rdb
 *
 * derived from WaveTest.java and expanded to support multiple wavelets
 *
 *     This program is intended to provide an example of the implementation
 * of wavelet transforms within the context of the Granite system.
 * 
 * The program accepts a command line argument that is an fdl description
 * of a DataSource. Portions of the program care about the dimensionality
 * of the DataSource and the number of attributes; others do not.
 * 
 * The program reads the data source does a wavelet on each attribute and
 * writes a new data source.
 * 
 * Command line options lets you select which wavelet to use; the intended
 * supported options are:
 *    -t h  Haar (default)
 *    -t d4 Daubechies 4
 *    -t d6 Daubechies 6
 *    -t d8 Daubechies 8
 *
 * The program also writes out elapsed time for each operation, but Java 
 * timing is tricky and timing I/O is even more problematic. There is a lot 
 * of initial overhead, so if you want to test a bunch of things in one run, 
 * don't trust the first test you do.
 * 
 * @author rdb
 */

package wavelet;

import java.util.ArrayList;
import java.io.*;
import edu.unh.sdb.common.*;
import edu.unh.sdb.datasource.*;

public class WaveTest
   {
      private static String     inName  = null;    // name of input descr file
      private static int[]      boundsArgs = null; // if not null, do subset
      
      private static String     waveOption = "-h"; // default code to run
      private static boolean    useDB = false;
      private static boolean    verbose = false;
      
      //------------------------------- main ---------------------------------
      public static void main( String[] args )
      {
         DataSource   ds = null;  // the original data source
         System.out.print( "WaveTest invoked with : " );
         for ( int i = 0; i < args.length; i++ )
            System.out.print( " " + args[ i ] );
         System.out.println();
         readArgs( args );
         if ( WaveTest.boundsArgs != null ) // if bounds specified do DB tests
            WaveTest.useDB = true;
         
         // The user has some control over what happens with errors discovered
         //   during execution. The Log.init method below says to save error
         //   error messages in a file called "Granite-errors" and to also send
         //   them to standard error.
         Log.init( "Granite-errors", Log.LogToFile | Log.LogToStdErr );
         
         // Error reporting levels include AlwaysThrow, ThrowFatalOnly, 
         // NeverThrow. It's probably best to generate errors as much as 
         // possible to be sure that you don't "miss" important problems.
         SDBError.setErrorReportingLevel( SDBError.AlwaysThrow );
         
         // Using the "Factory Method" design pattern, we let Granite figure out
         //   exactly what kind of DataSource is needed. The second name is
         //   arbitrary and is mostly for debugging and error messages.
         ds = openDataSource( "WaveTest", WaveTest.inName );  
         ds.activate();       // "activate" is equivalent to a file "open"
         DataCollection waveletSrc = ds;
         
         if ( WaveTest.useDB )
         {
            ISBounds selection;
            if ( WaveTest.boundsArgs == null )
               selection = ds.getBounds(); // get all of it
            else
            {
               if ( boundsArgs.length == ds.dim() )
                  selection = new ISBounds( boundsArgs );
               else if ( boundsArgs.length == 2 * ds.dim() )
               {
                  int[] lower = new int[ ds.dim() ];
                  int[] upper = new int[ ds.dim() ];
                  for ( int i = 0; i < ds.dim(); i++ )
                  {
                     lower[ i ] = boundsArgs[ i ];
                     upper[ i ] = boundsArgs[ i + ds.dim() ];
                  }
                  selection = new ISBounds( lower, upper );
               }
               else
               {
                  System.out.println( "Error: bounds argument must be same"
                                     + " as DataSource dimension (" + ds.dim()
                                     + ") or 2 times that. bounds length is: " 
                                     + boundsArgs.length );
                  System.out.println( "Using entire DataSource.");
                  selection = ds.getBounds();
               }
            }
            waveletSrc = ds.subblock( selection );
         }
         
         //+++++++++++++++++ Block and point iteration examples +++++++++++++++
         // Block iteration. Given 2n x 2n x 2n DataSource, produce n x n x n
         
         for ( int attr = 0; attr < ds.getNumAttributes(); attr++ )
            doWavelet( waveletSrc, attr, waveOption );
      }
      
      //---------------------- openDataSource ----------------------------
      // a little utility function to read the descriptor file.
      //
      private static DataSource openDataSource( String dsName, String fileName )
      {
         DataSource ds = null;
         try 
         {
            ds = DataSource.create( dsName, WaveTest.inName );
         } catch ( Exception ex )
         {
            System.err.println( "Open error: " + ex.getMessage() );
            System.exit( -1 );
         }
         return ds;
      }
      
      //----------------- doWavelet ----------------------------
      // A driver program to setup calls to the real Wavelet functions and do
      // some output.
      // 
      static void doWavelet( DataCollection highRes, int attr, String waveType )
      {
         Wavelet wave = null;
         long  start = 0;   // start time
         float elapsedSecs = 0.0f;
         
         
         System.out.println( "++++++++++++++++++++" + waveType 
                             + " Wavelet  ++++++++++++++++");
         System.out.println( "---------on attribute: " + attr );
         DataBlock lowRes = null;
         
         if ( verbose )
            printBlock( highRes.subblock( 
                                         new RecordSpec( new int[] { 0 }) ), 
                       "======== Input data ===========" );
         if ( waveType.equals( "h" ))
            wave = new HaarWavelet( highRes.dim() );
         else if ( waveType.equals( "d4" ))
            wave = new D4Wavelet( highRes.dim() );
         else
         {
            System.err.println( "Wave option not (yet) supported: " +
                    waveType );
            System.exit( -1 );
         }
            
         System.gc();
         start = System.currentTimeMillis();      
         
         lowRes = wave.doSummary( highRes, attr );
         
         System.out.println( "Wavelet time  (secs) = " + elapsedSecs );
         
         printBlock( lowRes, "========== Output data ===========" );
      }
      
      //------------------------ wavelet1D -----------------------------
      /** Shows basic access to 1D DataSources
       * The wavelet method can also handle 1D wavelets, but this method shows
       * a different approach to extracting and working with the data.
       */
      static DataBlock wavelet1D( DataCollection inData, int attr )
      {
         // output bounds will be half the size of the input
         int[]  outSize = new int[] { inData.getBounds().getDimension(0) / 2 };
         
         ISBounds  outBounds = new ISBounds( outSize );
         DataBlock outBlock  = inData.createDataBlock( outBounds, 
                                                      new RecordSpec( new int[] { attr }));
         
         // For the 1D version, we'll pull out the array, do the math
         //   and stuff the new array back into the output block in one shot.
         
         float[] outVals = new float [outBlock.volume() ];
         float[] inVals  = inData.getFloats( attr );
         
         // the Haar wavelet transform averages successive pairs of values
         for ( int i = 0; i < outVals.length; i++ )
            outVals[ i ] = ( inVals[ 2*i ] + inVals[ 2*i+1 ] )/ 2.0f;
         
         outBlock.setFloats( 0, outVals );
         return outBlock;
      }
      
      //------------------------ wavelet -----------------------------
      /**
       * Shows basic block access and block iteration over a data set.
       * Note that this code can do a Haar wavelet on any dimensionality,
       * with relatively simple code.
       */
      static DataBlock wavelet( DataCollection inData, int attr )
      {
         // First get the bounds of the input data; need it several places.
         // Notice that we use the "Reference" version to get the Bounds rather
         // than the copy version, getBounds(). 
         // Since we're not going to change the values, this is more efficient.
         
         ISBounds inBounds = inData.getBounds();
         // need output bounds to be 1/2 size of the input in all dimensions
         // first get the sizes of the input 
         int[]  outSizes = inBounds.getDimensionsArray(); 
         
         // now halve them -- for odd-length input, we'll just lose the end, 
         //        but thats ok
         for ( int i = 0; i < inData.dim(); i++ )    
            outSizes[i] = outSizes[i] / 2;
         
         // now make a bounds for the output
         ISBounds outBounds = new ISBounds( outSizes );
         
         // get a RecordSpec to extract just the right attribute
         RecordSpec rs = new RecordSpec( new int[] { attr });
         
         // get an output block with the right bounds and just one attribute
         DataBlock outBlock  = inData.createDataBlock( outBounds, rs );
         
         // For a 2D wavelet, extract a 2x2 block and average the 4 values.
         // To do the 3D wavelet, we need to extract 2x2x2 blocks from
         // the input and average all 8 values to get the output.
         // This code, however, is dimension-independent.
         
         int[]    waveBasis = new int[ inData.dim() ];
         for ( int i = 0; i < waveBasis.length; i++ )
            waveBasis[ i ] = 2;
         
         ISBounds waveBounds = new ISBounds( waveBasis );
         
         // And pre-allocated a DataBlock to take the 2x2x ... x 2 subblock
         DataBlock waveBlock = inData.createDataBlock( waveBounds, rs ); 
         
         // We need a block iterator that will go through the data space.      
         ISBoundsIterator inIter = new ISBoundsIterator( inBounds, waveBounds );
         
         // We also need an iterator over the output space, but this is a simple
         // iterator that just identifies one point at a time (not a block).
         ISIterator outIter = new ISIterator( outBounds );
         
         // pre-allocate an array to fill with the data
         float[] vals = new float[ waveBlock.volume() ];
         
         for ( ; inIter.valid(); inIter.next(), outIter.next() )
         {
            inData.subblock( waveBlock, inIter, rs );  // get the next block
            waveBlock.getFloats( vals );
            
            // now compute the average of the values in the block
            float sum = 0.0f;
            for ( int i = 0; i < vals.length; i++ )
               sum += vals[ i ];
            outBlock.setFloat( outIter, 0, sum/vals.length ); // store result
         }
         return outBlock;
      }
      
      //------------------------ printBlock -----------------------------
      // This method tries to do some basic formatting of the output. It works
      // ok for small files and dimensions of 3 or less.
      //
      static void printBlock( DataCollection inData, String header )
      {
         int      inDim = inData.dim();
         ISBounds inBounds = inData.getBounds();
         
         System.out.println( "++++++++++++++++++++++++++++++" + header 
                            + "++++++++++++++++++++++++++++++");
         if ( inDim == 1 )
            printRow( 0, inData.getFloats());
         else if ( inDim == 2 )
            printSlice( inData, 0 );
         else  // extract 3D slices one at a time
         {
            int sliceDim = inDim - 3;  // for 3D data set this is axis 0
            int numSlices = inBounds.getDimension( sliceDim );
            for ( int slice = 0; slice < numSlices; slice++ )
            {
               System.out.println( "------------------ slice "+  slice + 
                                  "----------------------------------------");
               printSlice( inData, slice );
            }
         }
         System.out.println( "++++++++++++++++++++++++++++++++++++++++++++++++" 
                            + "++++++++++++++++++++++++++++++");
      }
      
      //------------------------ printSlice -----------------------------
      static void printSlice( DataCollection inData, int whichSlice )
      {
         DataCollection printBlock = inData;
         
         int      inDim     = inData.dim();
         int      rowDim = inDim - 2;
         int      colDim = inDim - 1;
         int      sliceDim = inDim - 3;
         int      numRows;              // will be set to # rows to print
         float[]  rowVals = null;  // will be assign data for a row
         
         ISBounds inBnds    = inData.getBounds();
         int[]    sliceSizes = null;   // this array will first be lengths of a slice
         
         if ( inDim > 2 )
         {
            sliceSizes = new int [ inDim ];
            if ( inDim > 3 )
            {
               System.err.println( "printSlice: supports slices of 3D data " 
                                  + " only. Doing 1st 3d Hyperslice only.");
            }
            for ( int i = 0; i < rowDim; i++ )
               sliceSizes[ i ] = 1;
            
            sliceSizes[ rowDim ] = inBnds.getDimension( rowDim );
            sliceSizes[ colDim ] = inBnds.getDimension( colDim );
            ISBounds printBounds = new ISBounds( sliceSizes );
            
            // This printBounds actually identifies slice 0; we need to
            // change its bounds in the slice dimension to select the slice.
            try {
               printBounds.setLower( sliceDim, whichSlice );
               printBounds.setUpper( sliceDim, whichSlice );
            } catch ( Exception ex )
            { 
               System.err.println( "printSlice: bad slice. " + whichSlice );
            }
            printBlock = inData.subblock( printBounds );
         }
         
         ISBounds rowBounds = printBlock.getBounds();  // entire block at 1st
         
         if ( inDim == 1 )
            printRow( 0, printBlock.getFloats());
         else
         {
            numRows = printBlock.getBounds().getDimension( rowDim );
            for ( int r = 0; r < numRows; r++ )
            {
               rowBounds.setLower( rowDim, r ); // select just 1 row
               rowBounds.setUpper( rowDim, r ); 
               rowVals = printBlock.subblock( rowBounds ).getFloats( rowVals ); 
               // now print the data
               printRow( r, rowVals );
            }
         }
         System.out.println( "---------------------------------------------" );
      }
      //-------------- printRow  --------------------------------
      //
      static void printRow( int row, float[] vals )
      { 
         int      numValsPerLine = 20;
         int      numOnLine = 0;
         
         String   indent  = " ";
         String   indentIncr = "    ";
         
         for ( int c = 0; c < vals.length; c++ )
         {
            if ( numOnLine == 0 )
               System.out.print( "[" + row + "," + c + "]: " + indent );
            System.out.print( vals[ c ] + "   " );
            if ( ++numOnLine >= numValsPerLine )
            {
               System.out.println();
               numOnLine = 0;
               indent += indentIncr;
            }
            indent = " ";
         }
         if ( numOnLine != 0 )
            System.out.println();
      }
      
      //--------------- usage -----------------------------------
      static void usage()
      {
         PrintStream out = System.out; // should pass in output stream
         
         out.println( "Usage: java WaveTest fdl [-t opt] [bnds]" );
         out.println( " fdl  xfdl file describing the input DataSource" );
         out.println( " -t opt - Wavelet transform to use" );
         out.println( "      valid values of opt:" );
         out.println( "      h - Haar wavelet");
         out.println( "      d4 - Daubechies 4");
         out.println( "      d6 - Daubechies 4");
         out.println( "      d8 - Daubechies 4");
         out.println( " bnds  subblock of the DataSource as block input." );
         out.println( "       bnds must be either n ints or 2n ints where" );
         out.println( "       n is the dim of DataSource. Implies -db");      
      }
      
      //-------------- readArgs --------------------------------
      //
      static void readArgs( String[] args )
      {  
         if ( args.length == 0 )
         {
            usage();
            System.exit( 0 );
         }
         ArrayList<Integer> intArgs = new ArrayList<Integer>();
         for ( int iarg = 0; iarg < args.length; iarg++ )
         {
            String sarg = args[ iarg ];
            if ( sarg.equals( "-v" ))    // verbose flag
               WaveTest.verbose = true;
            else if ( sarg.equals( "-t" ))    // wavelet type flag
            {
               if ( iarg++ < args.length )
                  WaveTest.waveOption = args[ iarg ];
               else
                  System.err.println( "Error: -t needs an arg string." ); 
            }
            else if ( WaveTest.inName == null )  // not switch, check if ds
               WaveTest.inName = args[ iarg ];
            else // read bounds arguments
            {
               try 
               {
                  intArgs.add( new Integer( sarg ));
               }
               catch ( NumberFormatException nfe )
               {
                  System.err.println( "Error: Invalid argument: "
                                     + sarg + "  Expecting a number.");
               }
            }
         }
         if ( intArgs.size() > 0 )
         {
            WaveTest.boundsArgs = new int[ intArgs.size() ];
            for ( int i = 0; i < intArgs.size(); i++ )
               WaveTest.boundsArgs[i] = ((Integer)intArgs.get( i )).intValue();
         }
      }
   }
