package wavelet;

import selection.*;
import java.io.*;
import edu.unh.sdb.common.*;
import edu.unh.sdb.datasource.*;
import java.util.ArrayList;
import java.util.List;
import java.util.ListIterator;


/**
 * @author Sam Vohr
 *
 * This is a simple program using granite that creates smaller data sets
 * from a larger one using wavelets. 
 * 
 * Edit history 
 * 12/19/11 rdb: untyped ArrayList: errorFiles; its ArrayList<ErrorWriter> now.
 *
 */
public class WaveletDecomposer
{

    private int resolutions;
    private DataSource source;

    private FDLDescriptor fdl;
    private String sourceFilename;
    
    /**
     * Flag for suppressing the generation of FDL meta data files for the lower
     * resolutions and their error.
     */
    private boolean dataOnly = true;

    public WaveletDecomposer () 
    {

    }

    /**
     * Constructs an instance of the class using an FDL filename
     * and a number of resolutions to generate.
     * 
     * @param srcFile the filename of the FDL file for the orignal data
     * @param res the number of lower resolutions to generate
     */
    public WaveletDecomposer (String srcFile, int res)
    {
        this.sourceFilename = srcFile;

        this.fdl = FDLDescriptor.create( this.sourceFilename );
        this.source = openDataSource ( "res-zero", fdl ); 
        this.source.activate();

        this.resolutions = res;
    }
  
    /**
     * Prints the usage information for the main function.
     */
    private static void printUsage( )
    {
        System.err.println(
            "USAGE\n java selection.WaveletDecomposer [OPTIONS] FILE\n" +
            "    -n NUMBER_OF_RESOLUTIONS\n" +
            "        Default is 1\n\n" +
            "    --binary WIDTH HEIGHT DEPTH\n" +
            "        Specified FILE is binary, with size WIDTHxHEIGHTxDEPTH\n"+
            "        Byte order is little endian.\n\n"+
            "    --maxabs\n" +
            "        Use max absolute error (default).\n" +
            "    --snr\n" +
            "        Use Signal-to-Noise ratio for error.\n" +
            "    FILE is fdl (xml) input file, unless --binary is used.\n\n" +
            "    NOTE: CLASSPATH must include '.' and 'granite.jar'\n\n");
    }

    /**
     * The main function for the stand-alone decimation utility. 
     * 
     * @param args the number of resolutions and the FDL file to read from.
     *
     */
    public static void main( String[] args )
    {
        String fdlFileName = null;
        int numOfRes = 1;
        boolean keepFDL = true;

        try {
            // get the file to be read.
            if ( args.length > 0 ) {
                fdlFileName = args[args.length-1];

                for(int i=0; i<args.length; i++) { 
                    if(args[i].equals("-n")) {
                        numOfRes = Integer.parseInt(args[i+1]);
                    }

                    if(args[i].equals("--binary")) {
                        int width = Integer.parseInt(args[i+1]);
                        int height = Integer.parseInt(args[i+2]);
                        int depth = Integer.parseInt(args[i+3]);
                        String binfilename = args[args.length-1];

                        //System.err.printf("got size='%d %d %d'\n",
                                                    //width,height,depth);
                    
                        fdlFileName = "tmp.fdl";
                        PrintWriter pw = new PrintWriter(fdlFileName);

                        pw.printf("<?xml version=\"1.0\" encoding=\"UTF-" +
                                  "8\"?> <!DOCTYPE FileDescriptor PUBLIC " +
                                  "\"-//SDB//DTD//EN\"  \"fdl.dtd\"> " +
                                  "<FileDescriptor fileName= \"%s\" " +
                                  "fileType=\"binary\" byteOrder=\"little\"" +
                                  " > <Bounds lower= \"0\"    upper= \"%d\"" +
                                  " /> <Bounds lower= \"0\"    upper= \"%d\"" +
                                  "/> <Bounds lower= \"0\"    upper= " +
                                  "\"%d\"/> <Field fieldName= \"temp\" " +
                                  "fieldType= \"float\"/> </FileDescriptor>", 
                                  binfilename, depth-1, height-1, width-1);
                        pw.flush();
                        pw.close();
                        keepFDL = false;
                    }
                }

            } else {
                System.err.println ("Missing Arguments!");
                printUsage();
                System.exit(1);
            }
        } 
        catch(Exception e) {
            System.err.printf("ERROR: '%s'\n", e);
        }
    
        //System.err.printf("fdlFileName='%s'\n",fdlFileName);
        WaveletDecomposer d = new WaveletDecomposer (fdlFileName, numOfRes);
    
        d.reduce();

        if (!keepFDL)
            new File(fdlFileName).delete();
    }

    /**
     * Generates lower resolution data set by decimation using the 
     * previously established source and number of resolutions.
     *  
     */
    public void reduce ()
    { 
        
        // array of DataSetWriters
        DataSetWriter outFiles [] = new DataSetWriter [this.resolutions];
        List<ErrorWriter> errorFiles = new ArrayList<ErrorWriter> ();
        

        String outFileName = this.sourceFilename.substring(
                                0, this.sourceFilename.lastIndexOf('.') );
    
        // create files for the new resolutions
        for (int i=1, writer = 0; i <= this.resolutions; i++) {
            try {
                
                // build a new fdl decriptor for the new data set.
                FDLDescriptor fd = (FDLDescriptor)fdl.clone();
                fd.setBounds( 
                        shrinkByN( fdl.getBounds(), (int)Math.pow(2,i) ) );

                // create the data writer.
                String name = outFileName + "_" + i;
                outFiles[i-1] = new DataSetWriter (name, fd); 
                
                // create error files for each lower resolution
                for (int j=i; j <= this.resolutions; j++) {
                    ErrorWriter e;
                    
                    e = new MaxAbsErrorWriter (name, i, j, this.resolutions, 
                                fd.getBounds(), fdl);
                    e.suppressFDL(dataOnly); 
                    errorFiles.add(e);
                    
                    e = new SignalToNoiseErrorWriter (name, i, j, 
                                this.resolutions, fd.getBounds(), fdl);
                    e.suppressFDL(dataOnly); 
                    errorFiles.add(e);
                    
                    e = new AvgAbsErrorWriter (name, i, j, 
                                this.resolutions, fd.getBounds(), fdl);
                    e.suppressFDL(dataOnly); 
                    errorFiles.add(e);
                    
                    e = new StdDevErrorWriter (name, i, j, 
                                this.resolutions, fd.getBounds(), fdl);
                    e.suppressFDL(dataOnly); 
                    errorFiles.add(e);
                }
                
                outFiles[i-1].suppressFDL(dataOnly);
                
        
            } catch (Exception e) {
                System.err.println ("Cannot open output file.");
                System.exit(1);
            }
        }

        ISBounds bounds = this.source.getBounds(); // all of the data set.

        // the minimum number of slices required for the lowest resolution.
        int numSlices = (int)(Math.pow(2, this.resolutions)) + 1;
        int [] shape = {numSlices, 
                        bounds.getDimension(1), 
                        bounds.getDimension(2)};
        int [] gaps = {-1, 0, 0};
        

        // I think that this is the correct shape.
        System.err.println ("slice size " + bounds.getDimension(1) 
                + " x " + bounds.getDimension(2));

        ISBoundsIterator sliceIter = 
                new ISBoundsIterator (this.source.getBounds(), shape, gaps);


        for ( ; sliceIter.valid(); sliceIter.next() ) {

            ISBounds slices = sliceIter.copyBounds(); 
//          DataBlock sliceData = 
//              new BasicBlock (slice, this.source.getRecordDescriptor());
            
//          this.source.subblock(sliceData, slice);

            DataBlock sliceData = this.source.subblock(slices);

            // write the error values
            for ( ErrorWriter ew : errorFiles ) {
                try {
                    ew.write (sliceData);
                } catch (Exception e) {
                    System.err.println (
                          "Error in writing error data" );
                    e.printStackTrace(System.err);
                    System.err.println (e.toString());
                    System.exit(1);
                } 
            }
      

            for ( int j=1; j <= this.resolutions; j++ ) {

                // only build iterators for things we need
                // need every 2^j slice, where j is the level of resolution
                int step = (int)(Math.pow(2,j));
                

                int [] skip = {step,step,step};
                
                
                ISBounds sliceBounds = sliceData.getBounds();
                if ( sliceIter.isFullShape() ) {
                    // reduce the Z dimension by 1 to prevent overlap 
                    int [] lbound = sliceBounds.copyLowerArray();
                    int [] ubound = sliceBounds.copyUpperArray();
                    ubound[0] -= 1;
                    sliceBounds = new ISBounds (lbound,ubound);
                }
                    
                
                ISIterator iter = new ISIterator (sliceBounds, skip);
                Datum dat = sliceData.createDatum();

                // go through the data set, and output 
                for ( ; iter.valid(); iter.next() ) {
                    try {
                        sliceData.datum(dat, iter);
                        //System.err.println( dat.getFloat(0) );
                        outFiles[j-1].write( dat );

                    } catch (Exception e) {
                        System.err.println ("Error in writing resolution " + j);
                        System.exit(1);
                    }
                }
            }
        }


        // close everything
        try {
            for ( int i=0; i < this.resolutions; i++ ) {
                outFiles[i].close();
            }
            for ( ErrorWriter ew : errorFiles ) {
                ew.close();
            }
        } catch (Exception e) {
            System.err.println("unable to close file");
        }

        // done.

        return;
    }

    /**
     * utility function borrowed from BasicDemo 
     * @param dsName a String name for the data source
     * @param fdlD the FDLDescriptor built from the source's FDL file.
     */
    private DataSource openDataSource( String dsName, FDLDescriptor fdlD )
    {
        DataSource ds = null;
        try
        {
            ds = DataSource.create( dsName, fdlD );
        } catch ( Exception ex )
        {
            System.err.println( "Open error: " + ex.getMessage() );
            System.exit( -1 );
        }
        return ds;
    }


    /**
     * Returns an ISBounds with each dimension reduced by division
     * 
     * @param b an existing ISBounds
     * @param n the number of times to reduce each dimension
     * 
     */
    private ISBounds shrinkByN (ISBounds b, int n)
    {
        int [] dims = b.getDimensionsArray();

        for (int i=0; i < dims.length ; i++) 
            dims[i] /= n;

        return new ISBounds (dims);
    }
    
}

